'''
The contents of this file are focused on the plotting of the Data structure
in various projections and formats

These functions do this in whatever matplotlib instance you've got going on,
unless you toggle <show>
e.g. consider that below each function, I've added
    #show: if True, opens a window and shows the plot. Otherwise, adds it to
    #whatever matplotlib figure instance you have open
'''

#===============================================================================
# IMPORTS
#===============================================================================

import numpy as np
import matplotlib.pyplot as plt

#===============================================================================
# CONSTANTS & FLAGS
#===============================================================================

#used as the default background for fancy plots
default_fancy_background_url = 'https://upload.wikimedia.org/wikipedia/commons/thumb/4/4f/NGC_5170_HST_9766_R814B435.png/2560px-NGC_5170_HST_9766_R814B435.png'

#the licensing warning given when the user uses fancy plot
licensing_warning = "WARNING: The fancy_plot() routine uses an image that has been licensed by Rensselaer Polytechnic Institute (RPI). If you are not affiliated with RPI, then public usage of the default background image is not allowed. Please provide another URL for the background image with the `bg_url` keyword."

#===============================================================================
# FUNCTIONS
#===============================================================================

#generates a scatter plot of your Timestep object
def scatter(t, x, y, show=False, s=5.0, color='k', marker='o', **kwargs):
    #t: the Timestep object to plot
    #x: the x-axis parameter
    #y: the y-axis parameter

    plt.scatter(t[x], t[y], s=s, c=color, marker=marker, **kwargs)

    if show:
        plt.xlabel(x)
        plt.ylabel(y)
        plt.show()

#sticks a big fat red dot wherever the specific star(s) is, given an id(s)
#if vx and vy are specified, an arrow is drawn
def trace_particle(t, id, x, y, vx=None, vy=None, vscale=0.02, show=False, s=50., color='r', marker='o', **kwargs):
    #t (Timestep): the Timestep object that your particle is plotted over
    #id (int, can be array-like): the (list of) id(s) for your particle(s)
    #x (str): the x-axis parameter
    #y (str): the y-axis parameter
    #vx (str, optional): the x-axis arrow parameter
    #vy (str, optional): the y-axis arrow parameter
    #vscale (float, >0, optional): scales the arrow size
    #TODO: more customization of the arrow, probably

    #make sure id is always a numpy array of id values
    if not hasattr(id, "__len__"): #if true, id is not array-like:
        id = [id]
    else:
        if type(id)!=type(np.array([])):
            id = np.array(id)

    #get the indices of the particles from the id's
    #gotta be a 1-line way to do this with numpy, but I couldn't figure it out
    n = []
    for i in id:
        n.append(t['id'].index(i))

    #plot the particles
    for i in n:
        plt.scatter(t[x][i], t[y][i], s=s, c=color, marker=marker, **kwargs)
        if vx and vy:
            #TODO: figure out how to pass kwargs to the arrow and scatter separately
            plt.arrow(t[x][i], t[y][i], t[vx][i]*vscale, t[vy][i]*vscale, color=color, head_width=1)

    if show:
        plt.xlabel(x)
        plt.ylabel(y)
        plt.show()

#plots a histogram of the Timestep
#see np.hist for usage
def hist(t, x, show=False, range=None, bins=10, *args, **kwargs):
    #t (Timestep): the Timestep object being plotted
    #x (str): the axis parameter

    if range is None: #default range
        range = (min(t[x]), max(t[x]))

    h = plt.hist(t[x], range=range, bins=bins, *args, **kwargs)

    if show:
        plt.xlabel(x)
        plt.show()

    return h

def hist2d(t, x, y, show=False, *args, **kwargs):
    #t (Timestep): the Timestep object being plotted
    #x (str): the x-axis parameter
    #y (str): the y-axis parameter

    h = plt.hist2d(t[x], t[y], *args, **kwargs)

    if show:
        plt.xlabel(x)
        plt.ylabel(y)
        plt.show()

    return h

#-------------------------------------------------------------------------------
# fancy plots
#-------------------------------------------------------------------------------

#fancy plot is technically built off of plt.plot, not plt.scatter
#so the kwargs need to be for plot, not scatter.
def fancy_plot(t, ax1, ax2, show=False, ax=None, bg_url=default_fancy_background_url, no_axes=True, extent=(-100, 100, -100, 100), **kwargs):
    #t (Timestep): the Timestep that contains the data
    #ax1 (str): the x-axis, i.e. 'x' or 'z'
    #ax2 (str): the y-axis
    #show (bool): if True, shows the plot
    #ax (pyplot axis object): can pass an axis to place the fancy image in
    #    if None, then this routine just makes its own figure
    #bg_url (str): the url of an image online to use for the background image
    #    (by default this is an image that RPI has licensing for)
    #no_axes (bool): if True, removes the axes labels
    #extent (tuple of floats): (xmin, xmax, ymin, ymax) of plot bounds

    #out: returns the active pyplot axis, so that you can alter the image as needed

    #print the licensing warning if this is the first time that the user has made a fancy plot
    if bg_url == default_fancy_background_url:
        print(UserWarning(licensing_warning))

    # First set up the figure (if we have to)
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(12, 12))

    #set current axis
    plt.sca(ax)

    # read the image file in a numpy array
    a = plt.imread(bg_url)
    ax.imshow(a, extent=extent)

    #bottom right text
    if bg_url == default_fancy_background_url:
        ax.text(extent[0] + 0.05*(extent[1] - extent[0]), extent[2] + 0.05*(extent[3] - extent[2]), 'Background image credit: ESA/NASA', color='w',
                fontsize=20, zorder=1000, horizontalalignment='left', verticalalignment='bottom')

    #split up into dark and baryonic matter
    bm_indx = (t.typ == 0)
    dm_indx = (t.typ == 1)

    x_bm = t[ax1][bm_indx]
    y_bm = t[ax2][bm_indx]
    x_dm = t[ax1][dm_indx]
    y_dm = t[ax2][dm_indx]

    #plot baryonic matter
    if len(x_bm) > 0:
        bm_points, = ax.plot(x_bm, y_bm, color=(1, 1, 1, 1), marker='o', linestyle='', ms=1., zorder=100, **kwargs)
        bm_points2, = ax.plot(x_bm, y_bm, color=(0.8, 0.8, 0.8, 0.8), marker='o', linestyle='', ms=2., zorder=99, **kwargs)
        bm_points3, = ax.plot(x_bm, y_bm, color=(0.6, 0.6, 0.6, 0.6), marker='o', linestyle='', ms=4., zorder=98, **kwargs)
        bm_points4, = ax.plot(x_bm, y_bm, color=(0.4, 0.4, 0.4, 0.4), marker='o', linestyle='', ms=8., zorder=97, **kwargs)
        bm_points5, = ax.plot(x_bm, y_bm, color=(0.2, 0.2, 0.2, 0.2), marker='o', linestyle='', ms=16., zorder=96, **kwargs)

    #plot dark matter
    if len(x_dm) > 0:
        size_mult = 1.2
        dm_points, = ax.plot(x_dm, y_dm, color=(238/256,130/256,238/256, 1), marker='o', linestyle='', ms=1.*size_mult, zorder=95, **kwargs)
        dm_points2, = ax.plot(x_dm, y_dm, color=(221/256,160/256,221/256, 0.8), marker='o', linestyle='', ms=2.*size_mult, zorder=94, **kwargs)
        dm_points3, = ax.plot(x_dm, y_dm, color=(186/256,85/256,211/256, 0.6), marker='o', linestyle='', ms=4.*size_mult, zorder=93, **kwargs)
        dm_points4, = ax.plot(x_dm, y_dm, color=(148/256,0/256,211/256, 0.4), marker='o', linestyle='', ms=8.*size_mult, zorder=92, **kwargs)
        dm_points5, = ax.plot(x_dm, y_dm, color=(139/256,0/256,139/256, 0.2), marker='o', linestyle='', ms=16.*size_mult, zorder=91, **kwargs)

    ax.set_xlim(extent[0], extent[1])
    ax.set_ylim(extent[2], extent[3])

    if no_axes:
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)

    if show:
        plt.show()

    #return axis so you can manipulate it outside this function
    return ax

#-------------------------------------------------------------------------------



'''
This file is a place to store constants and other information
that is widely used throughout mwahpy.
'''

#===============================================================================
# IMPORTS
#===============================================================================

import astropy
import astropy.units as u
import matplotlib.pyplot as plt

import galpy
from galpy.potential import HernquistPotential
from galpy.potential import LogarithmicHaloPotential
from galpy.potential import MiyamotoNagaiPotential

from .mwahpy_glob import G

#===============================================================================
# CONSTANTS
#===============================================================================

m_bulge = 3.4e10*u.solMass #solar masses
m_disk = 1.0e11*u.solMass
v_halo = 74.61*u.km/u.s #km/s

pot_bulge = HernquistPotential(amp=2*m_bulge, a=0.7*u.kpc, ro=8., vo=220.)
pot_disk = MiyamotoNagaiPotential(amp=G*m_disk, a=6.5*u.kpc, b=0.26*u.kpc, ro=8., vo=220.)
pot_halo = LogarithmicHaloPotential(amp=2*v_halo**2, q=1., core=12.0*u.kpc, ro=8., vo=220.)

#this potential is from Newberg et al. 2010, Orphan Stream Model 5. It's basically Law 2005
mwahpy_default_pot = [pot_bulge, pot_disk, pot_halo]

energy_offset = -60000 #adjusts the energy to be consistent with Donlon et al. 2019

#===============================================================================
# FUNCTIONS
#===============================================================================

#produces a rotation curve for the given potential
def plot_potential(potential, Rrange=[0.01,10.]):
    fig = plt.figure(figsize=(12,8))
    potential.plotRotcurve(Rrange=Rrange, overplot=True)
    plt.show()



'''
The contents of this file are used to import data from milkyway .out files,
as well as write data out in a variety of formats
'''

#===============================================================================
# IMPORTS
#===============================================================================

import numpy as np
from pathlib import Path

from .flags import verbose, progress_bars
from .mwahpy_glob import file_len, progress_bar
from .timestep import Timestep
from .nbody import Nbody

#===============================================================================
# FUNCTIONS
#===============================================================================

#-------------------------------------------------------------------------------
# INPUT
#-------------------------------------------------------------------------------

def remove_header(f):
    #f (open file): the milkyway ".out" file

    comass = []
    comom = []

    #first 3 lines are junk
    for i in range(0, 3):
        f.readline()

    #next line has COM info
    line = f.readline()
    line = line.split(',')
    line[0] = line[0].strip('centerOfMass = ')
    line[3] = line[3].strip('centerOfMomentum = ')
    comass = [float(line[0]), float(line[1]), float(line[2])]
    comom = [float(line[3]), float(line[4]), float(line[5])]

    return comass, comom

#parses a milkyway ".out" file and returns a Timestep class structure
def read_output(f, start=None, stop=None):
    #f (str): the path of the milkyway ".out" file
    #start (optional): the line to actually start reading in data
    #stop (optional): the line to stop reading in data

    flen = 0 #allows this to be used outside of the next if statement's scope
    if progress_bars:
        flen = file_len(f)
        #properly adjust for length change based on start and stop points
        if start is not None:
            flen -= start
        if stop is not None:
            flen -= stop
    if verbose:
        print('\nReading in data from ' + str(f) + '...')

    f = open(f, 'r')

    #remove the header, get relevant info from header
    comass, comom = remove_header(f)

    #store the data here temporarily
    array_dict = {0:[], 1:[], 2:[], 3:[], 4:[], 5:[], 6:[], 7:[], 8:[], 9:[], 10:[], 11:[]}

    #place all the data from the file into the dictionary
    j = 0 #line num tracker

    #if start is specified,
    #skip lines until we're at the starting line
    if start is not None:
        while j < start:
            f.readline()
            j += 1

    j = 0 #reset line counter for progress bar
    for line in f:

        line = line.strip().split(',')
        i = 0
        while i < len(line) - 1: #this grabs l, b, r data even though that is calculated from x, y, z in the Timestep class implementation
                                 #it's mostly just for simplicity when reading in, although I guess it probably slows down the code somewhat
            array_dict[i].append(float(line[i]))
            i += 1
        j += 1

        if progress_bars:
            progress_bar(j, flen)

        #stop reading in lines if we're past the stop setting
        if stop is not None:
            if j >= stop:
                break
        else:
            continue  # only executed if the inner loop did NOT break
        break #only executed if the inner loop DID break

    #return the Timestep class using the array dictionary we built
    if verbose:
        print('\n'+ str(len(array_dict[1])) + ' objects read in')
        print('\rConverting data...', end='')
    d = Timestep(typ=array_dict[0], id_val=array_dict[1], x=array_dict[2], y=array_dict[3], z=array_dict[4], vx=array_dict[8], vy=array_dict[9], vz=array_dict[10], mass=array_dict[11], center_of_mass=comass, center_of_momentum=comom)
    if verbose:
        print('done')

    f.close()

    return d

#parses a milkyway ".in" file and returns a Timestep class structure
def read_input(f):
    #f (str): the path of the milkyway ".in" file

    if progress_bars:
        flen = file_len(f)
    if verbose:
        print('\nReading in data from ' + str(f) + '...')

    f = open(f, 'r')

    #remove the header
    f.readline()

    #store the data here temporarily
    array_dict = {0:[], 1:[], 2:[], 3:[], 4:[], 5:[], 6:[], 7:[], 8:[]}

    #place all the data from the file into the dictionary
    if progress_bars:
        j = 0
    for line in f:
        line = line.strip().split('\t')
        i = 0
        while i < len(line):
            array_dict[i].append(float(line[i]))
            i += 1
        if progress_bars:
            j += 1
            progress_bar(j, flen)

    #return the Timestep class using the array dictionary we built
    if verbose:
        print('\n'+ str(len(array_dict[1])) + ' objects read in')
        print('\rConverting data...', end='')
    d = Timestep(typ=array_dict[0], id_val=array_dict[1], x=array_dict[2], y=array_dict[3], z=array_dict[4], vx=array_dict[5], vy=array_dict[6], vz=array_dict[7], mass=array_dict[8], center_of_mass=[0,0,0], center_of_momentum=[0,0,0])
    if verbose:
        print('done')

    f.close()

    d.update(force=True) #this generates Comass and Comomentum, as well as the other bits
    #that you expect a Timestep to have initially when read in

    return d

#TODO: add nested progress bars
def read_folder(f, ts_scale=None):
    #f: the path to the folder that you want to create an Nbody structure out of
    #ts_scale: the scale of a single timestep in the Nbody sim

    if verbose:
        print('\nReading in data from directory ' + str(f) + '...')

    n = Nbody(ts_scale=ts_scale) #create the Nbody instance

    #iterate through the folder
    for i in Path(f).iterdir():

        t = read_output(str(i))
        time = int(str(i).split('/')[-1])
        n[time] = t #assign each timestep to the correct place in the Nbody

    return n

#-------------------------------------------------------------------------------
# OUTPUT
#-------------------------------------------------------------------------------

#parses a Timestep class object and outputs a file that can be read into a
#MilkyWay@home N-body simulation as the 'manual bodies' parameter
def make_nbody_input(t, f, recenter=True):
    #t (Timestep): the Timestep object that will be printed out
    #f (str): the path of the file that will be printed to
    if verbose:
        print('Writing Timestep as N-body input to '+f+'...')

    if recenter:
        t.recenter()

    f = open(f, 'w')
    f.write('#ignore\tid\tx\ty\tz\tvx\tvy\tvz\tm')

    i = 0

    while i < len(t):
        f.write('\n{}\t{}\t{:.6f}\t{:.6f}\t{:.6f}\t{:.6f}\t{:.6f}\t{:.6f}\t{} '.format(1, t.id[i], t.x[i], t.y[i], t.z[i], t.vx[i], t.vy[i], t.vz[i], t.mass[i]))
        #f.write('\n'+str(1)+' '+str(t.id[i])+' '+str(t.x[i])+' '+str(t.y[i])+' '+str(t.z[i])+' '+\
        #        str(t.vx[i])+' '+str(t.vy[i])+' '+str(t.vz[i])+' '+str(t.mass[i]))
        if progress_bars:
            progress_bar(i, len(t))
        i += 1

    f.write('\n')

    if verbose:
        print('\ndone')

#prints out a '.csv' file of the Timestep class structure
def make_csv(t, f_name):
    #t: the Timestep object being written out
    #f: the path for the output csv file

    f = open(f_name, 'w')

    #make the header
    #TODO: Print out COM's
    if verbose:
        print('Writing header...')
    header = ''
    for key in t:
        header += (key + ',')
    header += '\n'
    f.write(header)

    #iterate through the data and print each line
    i = 0
    if verbose:
        print('Printing data...')
    while i < len(t): #i avoided zip() here because it's messy to zip
    #                        like a billion different arrays, although zip()
    #                        is "better programming"
        if progress_bars:
            progress_bar(i, len(t))
        line = ''
        for key in t:
            line += (str(t[key][i]) + ',')
        line += '\n'
        f.write(line)
        i += 1

    print('Timestep output to ' + f_name)



import numpy as np
from numpy import linalg
import scipy as sc

from .mwahpy_glob import solar_ps

#TODO: Make sure all transformations have inverse transformations
#TODO: Make sure all transormations have unit tests
#TODO: (low priority) Allow certain quantities (like galactocentric phi/theta) to be expressed/input in radians
#TODO: (low priority) add left-handed versions to all galactic cartesian transformations

#===============================================================================
#HELPER FUNCTIONS
#===============================================================================
#These functions aren't meant to be accessed by outside files or by end users,
#and as such they are not well documented, named, or tested

def wrap_long(long, rad=False):

    if rad:
        long = long * 180/np.pi

    use_array = True
    if type(long) != type(np.array([])):
        use_array = False
        long = np.array([long])

    for i in range(len(long)):
        if long[i] < 0:
            long[i] += 360
        elif long[i] > 360:
            long[i] -= 360

    if rad:
        long = long * np.pi/180

    if not use_array:
        long = long[0]

    return long

#rotate the given data around the provided axis
#TODO: Allow array-like input
def rot_around_arb_axis(x, y, z, ux, uy, uz, theta):
    #TODO: Allow radians or degrees
    #x, y, z: 3D cartesian coordinates of the data
    #ux, uy, uz: 3d cartesian coordinates of the axis vector u = (ux, uy, uz)
    #theta: angle to rotate data counter-clockwise around axis (rad)

    #make sure that u is normalized
    norm_u = (ux**2 + uy**2 + uz**2)**0.5
    if round(norm_u, 8) != 1.:
        ux /= norm_u
        uy /= norm_u
        uz /= norm_u

    xyz = np.array([x, y, z])

    cos = np.cos(theta)
    sin = np.sin(theta)
    R = np.array([[cos+ux**2*(1-cos),    ux*uy*(1-cos)-uz*sin, ux*uz*(1-cos)+uy*sin],
                  [uy*ux*(1-cos)+uz*sin, cos+uy**2*(1-cos),    uy*uz*(1-cos)-ux*sin],
                  [uz*ux*(1-cos)-uy*sin, uz*uy*(1-cos)+ux*sin, cos+uz**2*(1-cos)   ]])

    xyz = np.matmul(R, xyz)
    x = xyz[0]
    y = xyz[1]
    z = xyz[2]

    return x, y, z

def long_lat_to_unit_vec(l, b, left_handed=False, rad=False):

    use_array = True
    if type(l) != type(np.array([])):
        use_array = False
        l = np.array([l])
        b = np.array([b])

    if not rad:
        l = l*np.pi/180
        b = b*np.pi/180

    if left_handed:
        #left-handed
        x = -1 * np.cos(l)*np.cos(b)
    else:
        #right-handed
        x = np.cos(l)*np.cos(b)

    y = np.sin(l)*np.cos(b)
    z = np.sin(b)

    if not use_array:
        x = x[0]
        y = y[0]
        z = z[0]

    return x, y, z

def comp_wise_dot(M, v, normalize=False):
    #performs a component-wise dot product for matrix M of vectors, with vector v
    #if normalize == True, normalize v and each component of M
    tol = 1e-6

    out = np.zeros(len(M))

    if normalize:
        v = v / linalg.norm(v)

    for i in range(len(M)):
        if normalize:
            if linalg.norm(M[i]) < tol: #required to handle singularities
                out[i] = 0
            else:
                out[i] = np.dot(M[i], v) / linalg.norm(M[i])
        else:
            out[i] = np.dot(M[i], v)

    return out

#===============================================================================
#FIRST ORDER COORDINATE/POSITION TRANSFORMATIONS
#===============================================================================

def cart_to_gal(x, y, z, left_handed=False):
    #get l, b, r (helio) from galactocentric X, Y, Z coords
    use_array = True
    if type(x) != type(np.array([])):
        use_array = False
        x = np.array([x])
        y = np.array([y])
        z = np.array([z])

    if left_handed:
        r = ((x+solar_ps.x)**2 + y**2 + z**2)**0.5
        l = np.arctan2(y,-1*(x-8))*180/np.pi
    else:
        r = ((x-solar_ps.x)**2 + y**2 + z**2)**0.5
        l = np.arctan2(y,(x+8))*180/np.pi
    b = np.arcsin(z/r)*180/np.pi

    if not use_array:
        l = l[0]
        b = b[0]
        r = r[0]

    return l, b, r

#-------------------------------------------------------------------------------

def gal_to_cart(l, b, r, left_handed=False, rad=False):

    use_array = True
    if type(l) != type(np.array([])):
        use_array = False
        l = np.array([l])
        b = np.array([b])
        r = np.array([r])

    if not rad:
        l = l*np.pi/180
        b = b*np.pi/180

    if left_handed:
        #left-handed
        x = -1*solar_ps.x - r*np.cos(l)*np.cos(b)
    else:
        #right-handed
        x = r*np.cos(l)*np.cos(b) + solar_ps.x

    y = r*np.sin(l)*np.cos(b)
    z = r*np.sin(b)

    if not use_array:
        x = x[0]
        y = y[0]
        z = z[0]

    return x, y, z

#-------------------------------------------------------------------------------

#input: x[kpc], y[kpc], z[kpc]
#output: Cylindrical coordinates (R, [kpc], Z [kpc], Phi [deg])
#phi = 0 when oriented along the positive x-axis
def cart_to_cyl(x, y, z):

    use_array = True
    if type(x) != type(np.array([])):
        use_array = False
        x = np.array([x])
        y = np.array([y])

    R = (x**2 + y**2)**0.5
    phi = np.arctan2(y, x)*180/np.pi

    if not use_array:
        R = R[0]
        phi = phi[0]

    return R, z, phi

#-------------------------------------------------------------------------------

def cyl_to_cart(R, z, phi):

    use_array = True
    if type(R) != type(np.array([])):
        use_array = False
        R = np.array([R])
        phi = np.array([phi])

    phi = phi*np.pi/180

    x = R*np.cos(phi)
    y = R*np.sin(phi)

    if not use_array:
        x = x[0]
        y = y[0]

    return x, y, z

#-------------------------------------------------------------------------------

def cart_to_sph(x, y, z):

    use_array = True
    if type(x) != type(np.array([])):
        use_array = False
        x = np.array([x])
        y = np.array([y])
        z = np.array([z])

    r = (x**2 + y**2 + z**2)**0.5

    phi = np.arctan2(y, x)*180/np.pi
    theta = np.arcsin(z/r)*180/np.pi

    if not use_array:
        phi = phi[0]
        theta = theta[0]
        r = r[0]

    return phi, theta, r

#-------------------------------------------------------------------------------

def sph_to_cart(phi, theta, r):

    use_array = True
    if type(phi) != type(np.array([])):
        use_array = False
        phi = np.array([phi])
        theta = np.array([theta])
        r = np.array([r])

    phi = phi*np.pi/180
    theta = theta*np.pi/180

    x = r*np.cos(phi)*np.cos(theta)
    y = r*np.sin(phi)*np.cos(theta)
    z = r*np.sin(theta)

    if not use_array:
        x = x[0]
        y = y[0]
        z = z[0]

    return x, y, z

#===============================================================================
#SECOND ORDER COORDINATE/POSITION TRANSFORMATIONS
#===============================================================================

def cyl_to_gal(R, z, phi):
    x, y, z = cyl_to_cart(R, z, phi)
    l, b, r = cart_to_gal(x, y, z)
    return l, b, r

#-------------------------------------------------------------------------------

def gal_to_cyl(l, b, r):
    x, y, z = gal_to_cart(l, b, r)
    R, z, phi = cart_to_cyl(x, y, z)
    return R, z, phi

#===============================================================================
#VELOCITY TRANSFORMATIONS
#===============================================================================

# Input: distance [kpc], radial velocity [km/s], RA/DEC [degrees], and pmRA/pmDEC [mas/yr]
# Returns: Galactic vx, vy, vz velocities [km/s]
# NOTE: pmRA = d/dt(RA) * cos(DEC)
# Arguments should be numpy arrays for most efficient usage
# Adapted from code written by Alan Pearl
def get_uvw(ra, dec, dist, rv, pmra, pmde):

    # Conversion from Equatorial (J2000) Cartesian to Galactic Cartesian
    EQ2GC = np.array( [[-0.05487572, -0.87343729, -0.48383453],
                      [  0.49410871, -0.44482923,  0.7469821 ],
                      [ -0.86766654, -0.19807649,  0.45598456]], dtype=np.float32 )

    ra_rad = ra * (np.pi/180.0)
    dec_rad = dec * (np.pi/180.0)

    sina = np.sin(ra_rad)
    cosa = np.cos(ra_rad)
    sind = np.sin(dec_rad)
    cosd = np.cos(dec_rad)

    vra =  4.741067035842384 * pmra * dist
    vdec = 4.741067035842384 * pmde * dist

    vx_Eq = rv * cosd * cosa   -   vdec * sind * cosa   -   vra * sina
    vy_Eq = rv * cosd * sina   -   vdec * sind * sina   +   vra * cosa
    vz_Eq = rv * sind          +   vdec * cosd

    vel_Eq = np.array([ vx_Eq, vy_Eq, vz_Eq ])
    U, V, W = np.dot(EQ2GC, vel_Eq)

    return U, V, W

#-------------------------------------------------------------------------------

def get_vxvyvz(ra, dec, dist, rv, pmra, pmde):

    U, V, W = get_uvw(ra, dec, dist, rv, pmra, pmde)

    # add Sun's velocity
    vx = U + solar_ps.vx
    vy = V + solar_ps.vy
    vz = W + solar_ps.vz

    return vx, vy, vz

#-------------------------------------------------------------------------------

#TODO: Do the linear algebra to avoid a inv calculation, since that's huge time waste
#solar reflex motion will be already removed if UVW are galactocentric
#inputs must be arrays, even if just of length 1
def get_rvpm(ra, dec, dist, U, V, W):

    use_array = True
    if type(ra) != type(np.array([])):
        use_array = False
        ra = np.array([ra])
        dec = np.array([dec])
        dist = np.array([dist])
        U = np.array([U])
        V = np.array([V])
        W = np.array([W])

    k = 4.74057
    rv = np.array([0.0]*len(ra))
    pmra = np.array([0.0]*len(ra))
    pmdec = np.array([0.0]*len(ra))

    i = 0
    while i < len(ra):

        ra_rad = ra[i] * np.pi/180
        dec_rad = dec[i] * np.pi/180

        T = np.array( [[-0.05487572, -0.87343729, -0.48383453],
                      [  0.49410871, -0.44482923,  0.7469821 ],
                      [ -0.86766654, -0.19807649,  0.45598456]], dtype=np.float32 )
        A = np.array([[np.cos(ra_rad)*np.cos(dec_rad), -1*np.sin(ra_rad), -1*np.cos(ra_rad)*np.sin(dec_rad)], [np.sin(ra_rad)*np.cos(dec_rad), np.cos(ra_rad), -1*np.sin(ra_rad)*np.sin(dec_rad)], [np.sin(dec_rad), 0, np.cos(dec_rad)]])
        B = np.matmul(T,A)
        B_inv = np.linalg.inv(B)

        uvw = np.array([[U[i]], [V[i]], [W[i]]])
        rvpm = np.matmul(B_inv, uvw)

        rv[i] = rvpm[0][0]
        pmra[i] = rvpm[1][0]/dist[i]/k
        pmdec[i] = rvpm[2][0]/dist[i]/k

        i += 1

    if not use_array:
        rv = rv[0]
        pmra = pmra[0]
        pmdec = pmdec[0]

    return rv, pmra, pmdec

#-------------------------------------------------------------------------------

#inputs must be arrays, even if just of length 1
def remove_sol_mot_from_pm(ra, dec, dist, pmra, pmdec):
    vx = np.array([-solar_ps.vx] * len(ra))
    vy = np.array([-solar_ps.vy] * len(ra))
    vz = np.array([-solar_ps.vz] * len(ra))

    rv, mura, mudec = get_rvpm(ra, dec, dist, vx, vy, vz)

    #don't directly modify pmra and pmdec
    pmra_new = pmra - mura
    pmdec_new = pmdec - mudec

    return pmra_new, pmdec_new

#-------------------------------------------------------------------------------

#does not work for arrays!
#TODO: make work with arrays
def get_uvw_errors(dist, ra, dec, pmra, pmdec, err_pmra, err_pmdec, err_rv, err_dist):
    #distance in pc
    k = 4.74057

    ra_rad = ra * np.pi/180
    dec_rad = dec * np.pi/180

    T = np.array([[-0.05487572,-0.87343729, -0.48383453], [0.49410871, -0.44482923,  0.7469821], [-0.86766654, -0.19807649,  0.45598456]])
    A = np.array([[np.cos(ra_rad)*np.cos(dec_rad), -1*np.sin(ra_rad), -1*np.cos(ra_rad)*np.sin(dec_rad)], [np.sin(ra_rad)*np.cos(dec_rad), np.cos(ra_rad), -1*np.sin(ra_rad)*np.sin(dec_rad)], [np.sin(dec_rad), 0, np.cos(dec_rad)]])
    B = np.matmul(T,A)
    C = B**2

    M = np.array([[err_rv**2],
                  [(k*dist)**2 * (err_pmra**2 + (pmra*err_dist/dist)**2)],
                  [(k*dist)**2 * (err_pmdec**2 + (pmdec*err_dist/dist)**2)]])
    N = 2*pmra*pmdec*k**2*err_dist**2*np.array([[B[0][1]*B[0][2]],
                                               [B[1][1]*B[1][2]],
                                               [B[2][1]*B[2][2]]])

    uvw_var = np.matmul(C,M) + N

    err_u = (uvw_var[0][0])**0.5
    err_v = (uvw_var[1][0])**0.5
    err_w = (uvw_var[2][0])**0.5

    return err_u, err_v, err_w

#remove solar reflex motion from line of sight velocity
def vlos_to_vgsr(l, b, vlos):
    #TODO: Allow ra, dec as inputs

    use_array = True
    if type(l) != type(np.array([])):
        use_array = False
        l = np.array([l])
        b = np.array([b])
        vlos = np.array([rv])

    l = l * np.pi/180
    b = b * np.pi/180

    vgsr = vlos + solar_ps.vx*np.cos(l)*np.cos(b) + solar_ps.vy*np.sin(l)*np.cos(b) + solar_ps.vz*np.sin(b)

    if not use_array:
        vgsr = vgsr[0]

    return vgsr

#add solar reflex motion back into GSR line of sight velocity
def vgsr_to_vlos(l, b, vgsr):
    #TODO: Allow ra, dec as inputs

    use_array = True
    if type(l) != type(np.array([])):
        use_array = False
        l = np.array([l])
        b = np.array([b])
        vgsr = np.array([vgsr])

    l = l * np.pi/180
    b = b * np.pi/180

    vlos = vgsr - solar_ps.vx*np.cos(l)*np.cos(b) - solar_ps.vy*np.sin(l)*np.cos(b) - solar_ps.vz*np.sin(b)

    if not use_array:
        vlos = vlos[0]

    return vlos

#=====================================
#MISC TOOLS
#=====================================

#-------------------------------------------------------------------------------

#this is more or less meant to be a helper function, I think.
#I won't add it to the documentation for that reason.
#borrowed from NewbyTools
def plane_dist(x,y,z, params):
    a,b,c = params
    return (a*x + b*y + c*z)

#-------------------------------------------------------------------------------

#x, y, z should be numpy arrays
#takes in galactocentric x, y, z data and outputs parameters for the best fit plane to those points incident the galactic center
#adapted from NewbyTools
def plane_OLS(x,y,z, print_distances=False):
    """ Solves for the best-fit plane to a set of x,y,z data using ordinary least-squares.
        Equation is of form z = Ax + By.
        DIFFERENT FROM NEWBY TOOLS plane_OLS() IN THAT WE CONSTRAIN THE PLANE THROUGH GALACTIC CENTER
        Output is normalized a,b,c of a plane of the form ax+by+cz=0"""
    A = np.array([x, y]).T
    B = z.T

    #solve Ax=B
    p = np.matmul(np.matmul(linalg.inv(np.matmul(A.T, A)), A.T), B.T) #uses left pseudo-inverse {(A^T * A)^-1 * A^T} due to system being overconstrained (A doesn't have a true inverse)
    params = [-float(p[0]), -float(p[1]), 1.0]  #c=1.0 by default
    bottom = np.sqrt(params[0]*params[0] + params[1]*params[1] + params[2]*params[2])
    for i in range(len(params)):  params[i] = params[i]/bottom
    print("# - Normalized best-fit plane parameters: {0}".format(params))
    if print_distances:
        for i in range(len(x)):
            print(plane_dist(x[i], y[i], z[i], params))
    return params

#-------------------------------------------------------------------------------

#get_plane_normal: [float, float, float, float] --> np.array([float, float, float])
#takes in parameters that define a plane in 3D and returns a normalized normal vector to that plane
def get_plane_normal(params):
    #params ([a, b, c]) corresponding to the equation for a plane ax + by + cz = 0
    #comes from the plan fitting method above

    #definition of a normal vector, given equation of a plane
    normal = np.array([params[0], params[1], params[2]])

    #normalize the normal vector
    len_normal = (normal[0]**2 + normal[1]**2 + normal[2]**2)**0.5
    normal = (normal[0]/len_normal, normal[1]/len_normal, normal[2]/len_normal)

    return normal

#===============================================================================
#NONSTANDARD/UNIQUE COORDINATE TRANFORMATIONS
#===============================================================================

#gal2plane: np.array(floats), np.array(floats), np.array(floats), (float, float, float), (float, float, float) --> np.array(floats), np.array(floats), np.array(floats)
#takes in galactic coordinates for a star(s) and returns their x,y,z coordinates with respect to a rotated plane with the normal vector provided
#Newby 2013 et al, appendix
def cart_to_plane(x, y, z, normal, point):

    #construct a 3D frame in the current long/lat frame
    xyz = np.array([x, y, z])

    #build the change of basis matrix
    r_pole = np.array(list(normal))
    r_origin = np.array(list(point))
    r_newy = np.cross(r_pole, r_origin)

    cob_mat = np.array([r_origin, r_newy, r_pole])

    #compute the new values
    xyz = np.matmul(cob_mat, xyz)
    x = xyz[0]
    y = xyz[1]
    z = xyz[2]

    return x, y, z

#gal2plane: np.array(floats), np.array(floats), np.array(floats), (float, float, float), (float, float, float) --> np.array(floats), np.array(floats), np.array(floats)
#takes in galactic coordinates for a star(s) and returns their x,y,z coordinates with respect to a rotated plane with the normal vector provided
#Newby 2013 et al, appendix
def gal_to_plane(l, b, d, normal, point):

    x, y, z = gal_to_cart(l, b, d)
    x, y, z = cart_to_plane(x, y, z, normal, point)

    return x, y, z

#-------------------------------------------------------------------------------

#kind of a helper function for the gal_to_lambet and cart_to_lambet functions,
#although it just finds longitude and latitude of any cartesian system
def cart_to_lonlat(x, y, z):
    Lam = np.arctan2(y, x)*180/np.pi #convert to degrees

    #correct Lam to be between 0 and 360 instead of -180 to 180
    i = 0
    while i < len(Lam):
        if Lam[i] < 0:
            Lam[i] += 360
        i += 1

    Bet = np.arcsin(z/(x**2 + y**2 + z**2)**0.5)*180/np.pi #convert to degrees

    return Lam, Bet

#-------------------------------------------------------------------------------

def gal_to_lambet(l, b, d, normal, point):

    x_prime, y_prime, z_prime = gal_to_plane(l, b, d, normal, point)
    Lam, Bet = cart_to_lonlat(x_prime, y_prime, z_prime)

    return Lam, Bet

#-------------------------------------------------------------------------------

#this may just go away. Not sure how useful it really is
def gal_to_lambet_galcentric(l, b, d, normal, point):

    galcx, galcy, galcz = gal_to_cart(l, b, d)
    x_prime, y_prime, z_prime = cart_to_plane(galcx, galcy, galcz, normal, point)
    Lam, Bet = cart_to_lonlat(x_prime, y_prime, z_prime)

    return Lam, Bet

#-------------------------------------------------------------------------------

def cart_to_lambet(x,y,z, normal, point):

    x_prime, y_prime, z_prime = cart_to_plane(x,y,z, normal=normal, point=point)
    Lam, Bet = cart_to_lonlat(x_prime, y_prime, z_prime)

    return Lam, Bet

#-------------------------------------------------------------------------------

def cart_to_sgr(x,y,z):

    x_prime = x_prime - solar_ps.x
    xyz = np.array([x, y, z])

    #Euler rotation matrix from Solar frame into Sgr
    phi = 183.8 * np.pi/180
    theta = 76.5 * np.pi/180
    psi = 194.1 * np.pi/180
    cphi = np.cos(phi)
    sphi = np.sin(phi)
    ctheta = np.cos(theta)
    stheta = np.sin(theta)
    cpsi= np.cos(psi)
    spsi= np.sin(psi)

    R = np.array([[cpsi*cphi-ctheta*sphi*spsi, cpsi*sphi+ctheta*cphi*spsi, spsi*stheta],
                   [-1*spsi*cphi-ctheta*sphi*cpsi, -1*spsi*sphi+ctheta*cphi*cpsi, cpsi*stheta],
                   [stheta*sphi, -1*stheta*cphi, ctheta]])

    xyz = np.matmul(R, xyz)
    x = xyz[0]
    y = xyz[1]
    z = xyz[2]

    Lam = np.arctan2(y, x)*180/np.pi
    Bet = np.arcsin(z/(x**2 + y**2 + z**2)**0.5)*180/np.pi

    return Lam, Bet

#-------------------------------------------------------------------------------

#gal2plane: np.array(floats), np.array(floats), np.array(floats), (float, float, float), (float, float, float) --> np.array(floats), np.array(floats)
#takes in galactic coordinates for a star(s) and returns their Lamba, Beta coordinates with respect to the Sgr stream plane
def gal_to_sgr(l, b):
    #l, b: Galactic coordinates (can be arrays)

    x = np.cos(l*np.pi/180)*np.cos(b*np.pi/180)
    y = np.sin(l*np.pi/180)*np.cos(b*np.pi/180)
    z = np.sin(b*np.pi/180)
    xyz = np.array([x, y, z])

    #Euler rotation matrix from Solar frame into Sgr
    phi = 183.8 * np.pi/180
    theta = 76.5 * np.pi/180
    psi = 194.1 * np.pi/180
    cphi = np.cos(phi)
    sphi = np.sin(phi)
    ctheta = np.cos(theta)
    stheta = np.sin(theta)
    cpsi= np.cos(psi)
    spsi= np.sin(psi)

    R = np.array([[   cpsi*cphi-ctheta*sphi*spsi,    cpsi*sphi+ctheta*cphi*spsi, spsi*stheta],
                  [-1*spsi*cphi-ctheta*sphi*cpsi, -1*spsi*sphi+ctheta*cphi*cpsi, cpsi*stheta],
                  [   stheta*sphi,                -1*stheta*cphi,                ctheta     ]])

    xyz = np.matmul(R, xyz)
    x = xyz[0]
    y = xyz[1]
    z = xyz[2]

    Lam = wrap_long(np.arctan2(y, x)*180/np.pi)
    Bet = np.arcsin(z/(x**2 + y**2 + z**2)**0.5)*180/np.pi

    return Lam, Bet

#-------------------------------------------------------------------------------

#gal2plane: np.array(floats), np.array(floats), np.array(floats), (float, float, float), (float, float, float) --> np.array(floats), np.array(floats)
#takes in galactic coordinates for a star(s) and returns their Lamba, Beta coordinates with respect to a rotated plane with the normal vector provided
def sgr_to_gal(Lam, Bet):
    #x, y, z: galactocentric x, y, z coordinates
    #normal: 3-vector providing the orientation of the plane to rotate into
    #point: 3-vector 'suggesting' the direction of the new x-axis

    x = np.cos(Lam*np.pi/180)*np.cos(Bet*np.pi/180)
    y = np.sin(Lam*np.pi/180)*np.cos(Bet*np.pi/180)
    z = np.sin(Bet*np.pi/180)
    xyz = np.array([x, y, z])

    #Euler rotation matrix from Solar frame into Sgr
    phi = 183.8 * np.pi/180
    theta = 76.5 * np.pi/180
    psi = 194.1 * np.pi/180
    cphi = np.cos(phi)
    sphi = np.sin(phi)
    ctheta = np.cos(theta)
    stheta = np.sin(theta)
    cpsi= np.cos(psi)
    spsi= np.sin(psi)

    R = np.array([[   cpsi*cphi-ctheta*sphi*spsi,    cpsi*sphi+ctheta*cphi*spsi, spsi*stheta],
                  [-1*spsi*cphi-ctheta*sphi*cpsi, -1*spsi*sphi+ctheta*cphi*cpsi, cpsi*stheta],
                  [   stheta*sphi,                -1*stheta*cphi,                ctheta     ]])
    R_inv = np.linalg.inv(R)

    xyz = np.matmul(R_inv, xyz)
    x = xyz[0]
    y = xyz[1]
    z = xyz[2]

    l = wrap_long(np.arctan2(y, x)*180/np.pi)
    b = np.arcsin(z/(x**2 + y**2 + z**2)**0.5)*180/np.pi

    return l, b

#-------------------------------------------------------------------------------

#sky_to_pole: array(float), array(float), tuple(float, float), tuple(float, float) -> array(float), array(float)
#rotate the positions on the sky (sky1, sky2) into a new frame determined by
#the pole of the new frame (pole1, pole2) and the origin of the new frame (origin1, origin2)
#The output is the longitude and latitude in the new coordinate frame

#NOTE: Inputs can be any spherical geometry, as long as the pole & origin arguments
#are in the same coordinate system as the (sky1, sky2) coordinates.
#e.g. if sky1 is a list of RAs and sky2 is a list of Decs, then
#the pole and origin arguments must also be specified in RA Dec.

def sky_to_pole(sky1, sky2, pole, origin, wrap=False, rad=False):
    #sky1, sky2: positions of the data on the sky (e.g. sky1 = array(RA), sky2 = array(Dec), etc.)
    #pole: position of the pole of the new coordinate system, tuple
    #origin: position of the origin of the new coordinate system, tuple
    #wrap: if True, Lam is constrained to only positive values. Otherwise, Lam is in [-180,180]
    #rad: is True, ALL inputs are in radians. ALL inputs should be in degrees if rad=False

    use_array = True
    if type(sky1) != type(np.array([])):
        use_array = False
        sky1 = np.array([sky1])
        sky2 = np.array([sky2])

    sky1 = sky1.copy() #fix aliasing
    sky2 = sky2.copy()
    pole1 = pole[0] #separate tuples for readability
    pole2 = pole[1] #and easy scalability
    origin1 = origin[0]
    origin2 = origin[1]

    if not rad:
        sky1 *= np.pi/180
        sky2 *= np.pi/180
        pole1 *= np.pi/180
        pole2 *= np.pi/180
        origin1 *= np.pi/180
        origin2 *= np.pi/180

    #construct a 3D frame in the current long/lat frame
    #all points have unit distance
    x = np.cos(sky1)*np.cos(sky2)
    y = np.sin(sky1)*np.cos(sky2)
    z = np.sin(sky2)
    xyz = np.array([x, y, z])

    #build the change of basis matrix
    r_pole = np.array([np.cos(pole1)*np.cos(pole2), np.sin(pole1)*np.cos(pole2), np.sin(pole2)])
    r_origin = np.array([np.cos(origin1)*np.cos(origin2), np.sin(origin1)*np.cos(origin2), np.sin(origin2)])
    r_newy = np.cross(r_pole, r_origin)

    cob_mat = np.array([r_origin, r_newy, r_pole])

    #compute the new values
    xyz = np.matmul(cob_mat, xyz)
    x = xyz[0]
    y = xyz[1]
    z = xyz[2]

    #calculate Lam, Bet in new frame
    Lam = np.arctan2(y, x)
    Bet = np.arcsin(z/(x**2 + y**2 + z**2)**0.5)

    #---------------------------

    if not rad:
        Lam *= 180/np.pi
        Bet *= 180/np.pi

    if wrap:
        Lam = wrap_long(Lam) #TODO: wrap_long should allow for radians

    if not use_array:
        Lam = Lam[0]
        Bet = Bet[0]

    return Lam, Bet

#===============================================================================
# RUNTIME
#===============================================================================



'''
This is a self-contained orbit fitting routine.

This orbit fitter is unique compared to other common orbit fitters in that it
uses a galactocentric generalized plane coordinate system when fitting data
'''

#TODO: Allow different coordinate systems input
#TODO: more settings put in through fit_orbit()

import numpy as np
import scipy.optimize as scopt
import matplotlib.pyplot as plt

from astropy import units as u
from galpy.orbit import Orbit

from .coords import get_plane_normal, plane_OLS, gal_to_lambet_galcentric
from .flags import verbose
from .pot import mwahpy_default_pot
from .orbit_fitter import make_orbit, get_OrbitData_from_orbit, get_point_list, OrbitData

'''
================================================================================
 FLAGS
================================================================================
'''

#-------------------------------------------------------------------------------
#DO NOT TOUCH
#Any routine meant to be used by an end user will configure this from input data
vx_flag = False
vy_flag = False
vz_flag = False
vgsr_flag = False
#-------------------------------------------------------------------------------

'''
================================================================================
 PARAMETERS FOR OPTIMIZATION
================================================================================
'''

#i think these can get mismatched between orbit_fitter.py and this file, so maybe
#we fix that if it's a problem

t_length = 0.5 #Gyr
resolution = 1000
ts = np.linspace(0, t_length, num=resolution)*u.Gyr

'''
================================================================================
 HELPER FUNCTIONS FOR OPTIMIZATION
================================================================================
'''

#take in data, orbit, and plane info: Output model data corresponding to each data point
def get_model_from_params(data, params, normal, point):

    #initialize the orbit we are fitting
    o = make_orbit(params)

    #we flip it around so that we are fitting both the forwards and the backwards orbit
    o_rev = o.flip()
    o_rev.integrate(ts, mwahpy_default_pot)

    data_orbit, data_orbit_rev = get_OrbitData_from_orbit(o, o_rev)
    data_orbit.calc_lam_bet(normal, point)
    data_orbit_rev.calc_lam_bet(normal, point)

    #grab full lists so that we can select the closest points once we get a list
    Lam = np.append(np.flip(data_orbit_rev.L), data_orbit.L)

    #get the list of points closest to each data point in Lambda
    point_list, costs = get_point_list(data.L, Lam)

    #grab the model points from the point list we grabbed
    Bet = np.append(np.flip(data_orbit_rev.B), data_orbit.B)
    B_model = np.array([Bet[p] for p in point_list]).flatten()

    D = np.append(np.flip(data_orbit_rev.d), data_orbit.d)
    D_model = np.array([D[p] for p in point_list]).flatten()

    if vx_flag:
        vx = np.append(np.flip(data_orbit_rev.vx), data_orbit.vx)
        vx_model = np.array([vx[p] for p in point_list]).flatten()
    else:
        vx_model = np.zeros(len(B_model))

    if vy_flag:
        vy = np.append(np.flip(data_orbit_rev.vy), data_orbit.vy)
        vy_model = np.array([vy[p] for p in point_list]).flatten()
    else:
        vy_model = np.zeros(len(B_model))

    if vz_flag:
        vz = np.append(np.flip(data_orbit_rev.vz), data_orbit.vz)
        vz_model = np.array([vz[p] for p in point_list]).flatten()
    else:
        vz_model = np.zeros(len(B_model))

    if vgsr_flag:
        vgsr = np.append(np.flip(data_orbit_rev.vlos), data_orbit.vlos)
        vgsr_model = np.array([vgsr[p] for p in point_list]).flatten()
    else:
        vgsr_model = np.zeros(len(B_model))

    return B_model, D_model, vx_model, vy_model, vz_model, vgsr_model, costs

#-------------------------------------------------------------------------------

#takes in the observed data and a test orbit and calculates the goodness-of-fit using a chi-squared method
def chi_squared(params, data=[], normal=(0, 0, 0), point=(1, 0, 0)):

    B_model, d_model, vx_model, vy_model, vz_model, vgsr_model, costs = get_model_from_params(data, params, normal, point) #get model data from orbit

    x2_B = sum(((B_model - data.B)/data.B_err)**2)
    x2_d = sum(((d_model - data.d)/data.d_err)**2)

    #this is the main chi squared that we are minimizing
    x2 = x2_B + x2_d + costs

    if vx_flag:
        x2_vx = sum(((vx_model - data.vx)/data.vx_err)**2)
        x2 += x2_vx

    if vy_flag:
        x2_vy = sum(((vy_model - data.vy)/data.vy_err)**2)
        x2 += x2_vy

    if vz_flag:
        x2_vz = sum(((vz_model - data.vz)/data.vz_err)**2)
        x2 += x2_vz

    if vgsr_flag:
        x2_vgsr = sum(((vgsr_model - data.vgsr)/data.vgsr_err)**2)
        x2 += x2_vgsr

    #there's a weird edge case where occasionally x2 is a short array of floats
    #this bit prevents scipy from throwing an error
    #no idea what causes that
    if type(x2) == type(np.array([])):
        x2 = x2[0]

    if verbose:
        print('GoF: ' + str(x2))

    return x2

#-------------------------------------------------------------------------------

#takes in data, then fits a Great Circle to that data and
#minimizes the chi_squared to fit an orbit to the data
def optimize(data_opt, max_it, bounds, guess, mode, **kwargs)

    #compute the preferred galactocentric spherical frame to fit this orbit in
    #(should be close to the orbital plane of the observed data)
    normal = get_plane_normal(plane_OLS(data_opt.x, data_opt.y, data_opt.z)) #get normal vector for fit Great Circle

    #construct the point vector (must be orthogonal to normal and be in the x-z plane)
    px = 1 / ((normal[0]/normal[2])**2 + 1)**0.5
    pz = -1 * px * (normal[0] / normal[2])
    point = (px, 0, pz)

    data_opt.calc_lam_bet(normal, point)

    '''
    ============================================================================
    DIFFERENTIAL EVOLUTION CONSTANTS
    ============================================================================
    '''
    #DO NOT TOUCH
    pop_size = 50 #10 times number of parameters
    diff_scaling_factor = 0.8
    crossover_rate = 0.9
    '''
    ============================================================================
    '''

    #the actual bulk of the optimization
    if mode == 'de': #do differential evolution

        if bounds is None:
            print(RuntimeWarning('Keyword `guess` was not provided in `fit_orbit()` (required for differential evolution): using default bounds of [(0, 360), (-90, 90), (0, 100), (-1000, 1000), (-1000, 1000), (-1000, 1000)].'))
            bounds = [(0, 360), (-90, 90), (0, 100), (-1000, 1000), (-1000, 1000), (-1000, 1000)]

        params = scopt.differential_evolution(chi_squared, bounds, args=(data_opt, normal, point,), maxiter=max_it, popsize=pop_size, mutation=diff_scaling_factor, recombination=crossover_rate, workers=-1, disp=not(verbose), **kwargs).x

    elif mode == 'gd': #do gradient descent

        if guess is None:
            print(RuntimeWarning('Keyword `guess` was not provided in `fit_orbit()` (required for gradient descent): using default guess of [0, 0, 0, 0, 0, 0].'))
            guess = [0, 0, 0, 0, 0, 0]

        #have to do it this way because you can't pass 'bounds' as a kwarg
        if bounds is not None:
            params = scopt.minimize(chi_squared, guess, args=(data_opt, normal, point,), bounds=bounds, options={'maxiter':max_it, 'disp':verbose}, **kwargs).x
        else:
            params = scopt.minimize(chi_squared, guess, args=(data_opt, normal, point,), options={'maxiter':max_it, 'disp':verbose}, **kwargs).x

    else:
        raise BaseException('`mode` for `fit_orbit()` must be either `de` for differential evolution, or `gd` for gradient descent.')

    x2 = chi_squared(params, data_opt, normal, point)

    return params, normal, point, x2

'''
================================================================================
 FUNCTIONS
================================================================================
'''

#pass in l, b, d values (and optional vx, vy, vz, vlos velocities)
#computes the best orbit fit to those parameters
#outputs the parameters for that best orbit fit
#also need to pass in errors for each value used (except l)

#additional kwargs are passed to scipy.optimize.differential_evolution()
def fit_orbit(l, b, b_err, d, d_err, vx=None, vy=None, vz=None, vgsr=None, \
              vx_err=None, vy_err=None, vz_err=None, vgsr_err=None, max_it=100, \
              bounds=None, guess=None, t_len=None, mode='de', **kwargs):

    #construct data
    #set proper flags based on input data
    if vx is not None:
        global vx_flag
        vx_flag = True
    if vy is not None:
        global vy_flag
        vy_flag = True
    if vz is not None:
        global vz_flag
        vz_flag = True
    if vgsr is not None:
        global vgsr_flag
        vgsr_flag = True

    #update t_length if necessary
    if t_len is not None:
        global t_length
        t_length = t_len
        global ts
        ts = np.linspace(0, t_length, num=resolution)*u.Gyr

    if verbose:
        print('Building orbit data...')

    data_opt = OrbitData(l, b, d, vx, vy, vz, vgsr, b_err, d_err, vx_err, vy_err, vz_err, vgsr_err)

    if verbose:
        print('===================================')
        print('Optimizing:')
        print('===================================')

    #optimization
    params, normal, point, x2 = optimize(data_opt, max_it, bounds, guess, mode, **kwargs)

    if verbose:
        print('===================================')
        print('Params: l, b, d, vx, vy, vz')
        print(params)
        print()
        print('Normal Vector:')
        print(normal)
        print()
        print('Point Vector:')
        print(point)
        print()
        print('Goodness-of-Fit:')
        print(x2)
        print('===================================')

    return params, normal, point, x2

'''
================================================================================
 PLOTTING
================================================================================
'''

def plot_orbit_lam_bet(L, B, params, normal, point):

    o = make_orbit(params)

    o_rev = o.flip()
    o_rev.integrate(ts, mwahpy_default_pot)

    data_orbit, data_orbit_rev = get_OrbitData_from_orbit(o, o_rev)
    data_orbit.calc_lam_bet(normal, point)
    data_orbit_rev.calc_lam_bet(normal, point)

    fig = plt.figure(figsize=(9, 6))

    plt.plot(data_orbit.L, data_orbit.B, c='b')
    plt.plot(data_orbit_rev.L, data_orbit_rev.B, c='r')
    plt.scatter(L, B, c='k')

    plt.xlim(0, 360)
    plt.ylim(-90, 90)
    plt.xlabel('$\\Lambda$')
    plt.ylabel('$\\beta$')

    plt.show()

def plot_orbit_gal(l, b, d, params):

    o = make_orbit(params)

    o_rev = o.flip()
    o_rev.integrate(ts, mwahpy_default_pot)

    data_orbit, data_orbit_rev = get_OrbitData_from_orbit(o, o_rev)

    fig = plt.figure(figsize=(18, 8))
    ax1 = fig.add_subplot(121)
    ax2 = fig.add_subplot(122)

    ax1.plot(data_orbit.l, data_orbit.b, c='b')
    ax1.plot(data_orbit_rev.l, data_orbit_rev.b, c='r')
    ax1.scatter(l, b, c='k')

    ax1.set_xlim(0, 360)
    ax1.set_ylim(-90, 90)
    ax1.set_xlabel('l')
    ax1.set_ylabel('b')

    ax2.plot(data_orbit.l, data_orbit.d, c='b')
    ax2.plot(data_orbit_rev.l, data_orbit_rev.d, c='r')
    ax2.scatter(l, d, c='k')

    ax2.set_xlim(0, 360)
    ax2.set_xlabel('l')
    ax2.set_ylabel('d (helio)')

    plt.show()

'''
================================================================================
 TESTING
================================================================================
'''

def test():
    test_o_l = 90
    test_o_b = 20
    test_o_d = 25
    test_o_vx = 150
    test_o_vy = -200
    test_o_vz = 100

    ts = np.linspace(0, 0.25, 1000)*u.Gyr
    sample = np.array([100, 250, 400, 500, 600, 750, 850])

    o = make_orbit([test_o_l, test_o_b, test_o_d, test_o_vx, test_o_vy, test_o_vz])

    l = np.take(o.ll(ts), sample)
    b = np.take(o.bb(ts), sample)
    d = np.take(o.dist(ts), sample)

    plt.scatter(l, b, s=5)
    plt.scatter(o.ll(ts), o.bb(ts), s=1)
    plt.show()

    plt.scatter(l, d, s=5)
    plt.scatter(o.ll(ts), o.dist(ts), s=1)
    plt.show()

    b_err = np.zeros(len(b)) + 0.05
    d_err = np.zeros(len(d)) + 1

    test_orbit_data = OrbitData(l, b, d, None, None, None, None, b_err, d_err, None, None, None, None)
    test_orbit_data.calc_lam_bet((0, 0, 1), (1, 0, 0))

    print('Goodness-of-Fit of actual values:')
    print(chi_squared([test_o_l, test_o_b, test_o_d, test_o_vx, test_o_vy, test_o_vz], data=test_orbit_data, normal=(0, 0, 1), point=(1, 0, 0)))

    params, normal, point, x2 = fit_orbit(l, b, b_err, d, d_err)
    L, B = gal_to_lambet_galcentric(l, b, d, normal, point)
    plot_orbit_lam_bet(L, B, params, normal, point)
    plot_orbit_gal(l, b, d, params)

'''
================================================================================
 RUNTIME
================================================================================
'''

if __name__ == "__main__":
    test()



'''
This is a self-contained orbit fitting routine.
'''

#TODO: Allow fitting on b, ra, or dec instead of l

import numpy as np
import scipy as sc
import scipy.optimize as scopt
import matplotlib.pyplot as plt

from astropy import units as u
from astropy.coordinates import SkyCoord

import galpy
from galpy.orbit import Orbit

from .flags import verbose
from .pot import mwahpy_default_pot
from .coords import cart_to_lambet

'''
================================================================================
 FLAGS
================================================================================
'''

#-------------------------------------------------------------------------------
#DO NOT TOUCH
#Any routine meant to be used by an end user will configure this from input data
vx_flag = False
vy_flag = False
vz_flag = False
vgsr_flag = False
#-------------------------------------------------------------------------------

'''
================================================================================
 PARAMETERS FOR OPTIMIZATION
================================================================================
'''
t_length = 0.5 #Gyr
resolution = 1000
ts = np.linspace(0, t_length, num=resolution)*u.Gyr

deg_sep_pun = 1000 #multiplier for every degree that the lambda fitting function is off
#this can be tweaked based on how many degrees of freedom you have - fewer DOF
#means punishment can/should be smaller

'''
================================================================================
 CLASSES
================================================================================
'''

class OrbitData():
    #all of these parameters can be np.arrays of floats
    def __init__(self, l, b, d, vx, vy, vz, vgsr, b_err, d_err, vx_err, vy_err, vz_err, vgsr_err):
        self.l = l
        self.b = b
        self.d = d #heliocentric distance
        self.vx = vx
        self.vy = vy
        self.vz = vz
        self.vgsr = vgsr

        self.b_err = b_err
        self.d_err = d_err
        self.vx_err = vx_err
        self.vy_err = vy_err
        self.vz_err = vz_err
        self.vgsr_err = vgsr_err

        self.x = self.d*np.cos(self.l*np.pi/180)*np.cos(self.b*np.pi/180) - 8
        self.y = self.d*np.sin(self.l*np.pi/180)*np.cos(self.b*np.pi/180)
        self.z = self.d*np.sin(self.b*np.pi/180)

        self.r = (self.x**2 + self.y**2 + self.z**2)**0.5

    #add the icrs converted coordinates to this data instance
    def icrs(self):
        s = SkyCoord(self.l, self.b, frame='galactic', unit=(u.deg, u.deg))
        s = s.transform_to('icrs')
        self.ra = s.ra
        self.dec = s.dec

    #compute the lat/long for an arbitrary plane, given by its normal vector and rough origin point. \
    #used in orbit_fitter_gc
    def calc_lam_bet(self, normal, point):
        #normal: the normal vector to the plane of the Great Circle we are estimating for the orbit
        #point: parameter for the axis generation of the Great Circle coordinates
        self.L, self.B = cart_to_lambet(self.x, self.y, self.z, normal, point)
        self.B_err = self.b_err #should be error propogated (Newby et al. 2013) but is not yet

'''
================================================================================
 HELPER FUNCTIONS FOR OPTIMIZATION
================================================================================
'''

#do this in one place so it's uniform across the file and easily changeable
def make_orbit(params, int_ts=None, pot=None):

    #negate the x velocity because galpy is in left-handed frame and we aren't barbarians
    o = Orbit(vxvv=[params[0]*u.deg, params[1]*u.deg, params[2]*u.kpc, -1*params[3]*u.km/u.s, params[4]*u.km/u.s, params[5]*u.km/u.s], uvw=True, lb=True, ro=8., vo=220., zo=0., solarmotion=[0, -220, 0])

    if int_ts: #use the global ts unless explicitly provided with a different ts
        o.integrate(int_ts, pot)
    else:
        o.integrate(ts, pot)

    return o

#-------------------------------------------------------------------------------

#this function just makes a few places in the code cleaner
def get_OrbitData_from_orbit(o, o_rev):
    #o: the orbit
    #o: the reversed orbit
    #both of these should be integrated prior to calling this function

    #don't think that the velocities need obs=[8., 0., etc] except the vlos (to make it vgsr)
    data_orbit = OrbitData(np.array(o.ll(ts)), np.array(o.bb(ts)), np.array(o.dist(ts)), np.array(o.vx(ts, obs=[8., 0., 0., 0., 0., 0.]))*-1, np.array(o.vy(ts, obs=[8., 0., 0., 0., 0., 0.])), np.array(o.vz(ts, obs=[8., 0., 0., 0., 0., 0.])), np.array(o.vlos(ts, obs=[8., 0., 0., 0., 0., 0.])), np.array([]), np.array([]), np.array([]), np.array([]), np.array([]), np.array([]))
    data_orbit_rev = OrbitData(np.array(o_rev.ll(ts)), np.array(o_rev.bb(ts)), np.array(o_rev.dist(ts)), np.array(o_rev.vx(ts, obs=[8., 0., 0., 0., 0., 0.])), np.array(o_rev.vy(ts, obs=[8., 0., 0., 0., 0., 0.]))*-1, np.array(o_rev.vz(ts, obs=[8., 0., 0., 0., 0., 0.]))*-1, np.array(o.vlos(ts, obs=[8., 0., 0., 0., 0., 0.]))*-1, np.array([]), np.array([]), np.array([]), np.array([]), np.array([]), np.array([]))

    return data_orbit, data_orbit_rev

#-------------------------------------------------------------------------------

#given the full list of Lambdas, outputs the indices of the points within that list closest to our data's Lambdas
#(while keeping in mind that it may wrap from 0 to 360 degrees and vice versa)
def get_point_list(vals, Lam):
    #vals: the Lambda values that you want to find the closest indices to
    #Lam: the array of Lambda values that you are searching through

    #---------------------------------------------------------------------------
    #grabs the index of the value in Lam with the closest value to val
    def get_closest_index(val, Lam):
        Lam = np.abs(Lam - val)
        m = np.min(Lam)

        try: #another race condition?
            ind = np.where(Lam == m)[0][0]
        except IndexError:
            ind = -1

        return ind, m*deg_sep_pun
    #---------------------------------------------------------------------------

    point_list = []
    Lam_list = []
    costs = 0

    for val in vals:

        #within that segment, find the index which produces the value closest to val
        point, c = get_closest_index(val, Lam)
        costs += c

        #toss it in the list
        point_list.append(point)
        Lam_list.append(Lam[point])

    return point_list, costs

#get_model_from_orbit: data, orbit, vector, vector -> list(int) x3
#take in data, orbit, and plane info: Output model data corresponding to each data point
def get_model_from_orbit(data, params, pot=None):
    #data: the data that the orbit is being fit to
    #o: the test orbit that we are calculating the goodness-of-fit of
    #normal: the normal vector to the plane of the Great Circle we are estimating for the orbit
    #point: parameter for the axis generation of the Great Circle coordinates

    #initialize the orbit we are fitting
    o = make_orbit(params, pot=pot)

    #we flip it around so that we are fitting both the forwards and the backwards orbit
    o_rev = o.flip()
    o_rev.integrate(ts, mwahpy_default_pot)

    data_orbit, data_orbit_rev = get_OrbitData_from_orbit(o, o_rev)

    #grab full lists so that we can select the closest points once we get a list
    Lam = np.append(np.flip(data_orbit_rev.l), data_orbit.l)

    #get the list of points closest to each data point in Lambda
    point_list, costs = get_point_list(data.l, Lam)

    #grab the model points from the point list we grabbed
    Bet = np.append(np.flip(data_orbit_rev.b), data_orbit.b)
    B_model = np.array([Bet[p] for p in point_list]).flatten()

    D = np.append(np.flip(data_orbit_rev.d), data_orbit.d)
    D_model = np.array([D[p] for p in point_list]).flatten()

    if vx_flag:
        vx = np.append(np.flip(data_orbit_rev.vx), data_orbit.vx)
        vx_model = np.array([vx[p] for p in point_list]).flatten()
    else:
        vx_model = np.zeros(len(B_model))

    if vy_flag:
        vy = np.append(np.flip(data_orbit_rev.vy), data_orbit.vy)
        vy_model = np.array([vy[p] for p in point_list]).flatten()
    else:
        vy_model = np.zeros(len(B_model))

    if vz_flag:
        vz = np.append(np.flip(data_orbit_rev.vz), data_orbit.vz)
        vz_model = np.array([vz[p] for p in point_list]).flatten()
    else:
        vz_model = np.zeros(len(B_model))

    if vgsr_flag:
        vgsr = np.append(np.flip(data_orbit_rev.vgsr), data_orbit.vgsr)
        vgsr_model = np.array([vgsr[p] for p in point_list]).flatten()
    else:
        vgsr_model = np.zeros(len(B_model))

    return B_model, D_model, vx_model, vy_model, vz_model, vgsr_model, costs

#chi_squared: data, galpy.Orbit() --> float
#takes in the observed data and a test orbit and calculates the goodness-of-fit using a chi-squared method
def chi_squared(params, data=[], pot=None):
    #data: the data that the orbit is being fit to
    #o: the test orbit that we are calculating the goodness-of-fit of
    #normal: the normal vector to the plane of the Great Circle we are estimating for the orbit
    #point: parameter for the axis generation of the Great Circle coordinates

    B_model, d_model, vx_model, vy_model, vz_model, vgsr_model, costs = get_model_from_orbit(data, params, pot=pot) #get model data from orbit

    #B_model sometimes has different length than data.b, no idea why
    #I think it might be a race condition
    #this keeps the script running and tells the optimizer that the parameters are bad
    if len(B_model) != len(data.b):
        return 1e10

    x2_B = sum(((B_model - data.b)/data.b_err)**2)
    x2_d = sum(((d_model - data.d)/data.d_err)**2)

    #this is the main chi squared that we are minimizing
    #it's not a "chi-squared" anymore, but multiplying by a constant seems dumb
    x2 = x2_B + x2_d + costs #Willett et al. 2009, give or take

    if vx_flag:
        x2_vx = sum(((vx_model - data.vx)/data.vx_err)**2)
        x2 += x2_vx

    if vy_flag:
        x2_vy = sum(((vy_model - data.vy)/data.vy_err)**2)
        x2 += x2_vy

    if vz_flag:
        x2_vz = sum(((vz_model - data.vz)/data.vz_err)**2)
        x2 += x2_vz

    if vgsr_flag:
        x2_vgsr = sum(((vgsr_model - data.vgsr)/data.vgsr_err)**2)
        x2 += x2_vgsr

    '''
    #get normalization factor
    N = len(data.l) #number of data points
    n = 5 #number of parameters
    eta = N - n - 1 #normalizing parameter
    if eta <= 0:
        eta = 1 #if you use fewer data points than needed to constrain the problem, then this will still work but it won't be normalized correctly

    x2 = (1/eta) * (x2_B + x2_d + x2_vx + x2_vy + x2_vz + x2_vgsr) + costs #Willett et al. 2009, give or take
    '''

    #there's a weird edge case where occasionally x2 is a short array of floats
    #this bit prevents scipy from throwing an error
    #no idea what causes that
    if type(x2) == type(np.array([])):
        x2 = x2[0]

    if verbose:
        print('GoF: ' + str(x2))

    return x2

#optimize: data -> [float, float, float, float, float], (float, float, float), (float, float, float)
#takes in data, then fits a Great Circle to that data and minimizes the chi_squared to fit an orbit to the data
def optimize(data_opt, max_it, bounds, guess, mode, method='Nelder-Mead', pot=None, **kwargs):

    '''
    ============================================================================
    DIFFERENTIAL EVOLUTION CONSTANTS
    ============================================================================
    '''
    #DO NOT TOUCH
    pop_size = 50 #10 times number of parameters
    diff_scaling_factor = 0.8
    crossover_rate = 0.9
    '''
    ============================================================================
    '''

    if mode == 'de': #do differential evolution

        if bounds is None:
            print(RuntimeWarning('Keyword `bounds` was not provided in `fit_orbit()` (required for differential evolution): using default bounds of [(0, 360), (-90, 90), (0, 100), (-500, 500), (-500, 500), (-500, 500)].'))
            bounds = [(0, 360), (-90, 90), (0, 100), (-500, 500), (-500, 500), (-500, 500)]

        params = scopt.differential_evolution(chi_squared, bounds, args=(data_opt, pot,), maxiter=max_it, popsize=pop_size, mutation=diff_scaling_factor, recombination=crossover_rate, workers=-1, disp=not(verbose), **kwargs).x

    elif mode == 'gd': #do gradient descent

        if guess is None:
            print(RuntimeWarning('Keyword `guess` was not provided in `fit_orbit()` (required for gradient descent): using default guess of [0, 0, 0, 0, 0, 0].'))
            guess = [0, 0, 0, 0, 0, 0]

        #have to do it this way because you can't pass 'bounds' as a kwarg
        if bounds is not None:
            params = scopt.minimize(chi_squared, guess, args=(data_opt, pot,), bounds=bounds, options={'maxiter':max_it, 'disp':verbose}, method=method, **kwargs).x
        else:
            params = scopt.minimize(chi_squared, guess, args=(data_opt, pot,), options={'maxiter':max_it, 'disp':verbose}, method=method, **kwargs).x

    else:
        raise BaseException('`mode` for `fit_orbit()` must be either `de` for differential evolution, or `gd` for gradient descent.')

    x2 = chi_squared(params, data=data_opt, pot=pot)

    return params, x2

'''
================================================================================
 FUNCTIONS
================================================================================
'''

def fit_orbit(l, b, b_err, d, d_err, vx=None, vy=None, vz=None, vgsr=None, \
                 vx_err=None, vy_err=None, vz_err=None, vgsr_err=None, max_it=100, \
                 bounds=None, guess=None, t_len=None, mode='de', pot=None, **kwargs):

    if pot is None:
        print(RuntimeWarning("Didn't provide potential for constructing orbits! Using mwahpy_default_pot for potential model."))
        pot = mwahpy_default_pot

    #construct data
    #set proper flags based on input data
    if vx is not None:
        global vx_flag
        vx_flag = True
    if vy is not None:
        global vy_flag
        vy_flag = True
    if vz is not None:
        global vz_flag
        vz_flag = True
    if vgsr is not None:
        global vgsr_flag
        vgsr_flag = True

    #update t_length if necessary
    if t_len is not None:
        global t_length
        t_length = t_len
        global ts
        ts = np.linspace(0, t_length, num=resolution)*u.Gyr

    if verbose:
        print('Building orbit data...')

    data_opt = OrbitData(l, b, d, vx, vy, vz, vgsr, b_err, d_err, vx_err, vy_err, vz_err, vgsr_err)

    if verbose:
        print('===================================')
        print('Optimizing:')
        print('===================================')

    #optimization
    params, x2 = optimize(data_opt, max_it, bounds, guess, mode, pot=pot, **kwargs)

    print('===================================')
    print('Params: l, b, d, vx, vy, vz')
    print(params)
    print()
    print('Goodness-of-Fit:')
    print(x2)
    print('===================================')

    return params, x2

'''
================================================================================
 PLOTTING
================================================================================
'''

def plot_orbit_gal(l, b, d, params, pot=None):

    o = make_orbit(params, pot=pot)

    o_rev = o.flip()
    o_rev.integrate(ts, mwahpy_default_pot)

    data_orbit, data_orbit_rev = get_OrbitData_from_orbit(o, o_rev)

    fig = plt.figure(figsize=(18, 8))
    ax1 = fig.add_subplot(121)
    ax2 = fig.add_subplot(122)

    ax1.plot(data_orbit.l, data_orbit.b, c='b')
    ax1.plot(data_orbit_rev.l, data_orbit_rev.b, c='r')
    ax1.scatter(l, b, c='k')

    ax1.set_xlim(0, 360)
    ax1.set_ylim(-90, 90)
    ax1.set_xlabel('l')
    ax1.set_ylabel('b')

    ax2.plot(data_orbit.l, data_orbit.d, c='b')
    ax2.plot(data_orbit_rev.l, data_orbit_rev.d, c='r')
    ax2.scatter(l, d, c='k')

    ax2.set_xlim(0, 360)
    ax2.set_xlabel('l')
    ax2.set_ylabel('d (helio)')

    plt.show()

def plot_orbit_icrs(l, b, d, params, pot=None):

    s = SkyCoord(l, b, frame='galactic', unit=(u.deg, u.deg))
    s = s.transform_to('icrs')
    ra = s.ra
    dec = s.dec

    o = make_orbit(params, pot=pot)

    o_rev = o.flip()
    o_rev.integrate(ts, mwahpy_default_pot)

    data_orbit, data_orbit_rev = get_OrbitData_from_orbit(o, o_rev)
    data_orbit.icrs()
    data_orbit_rev.icrs()

    fig = plt.figure(figsize=(18, 8))
    ax1 = fig.add_subplot(121)
    ax2 = fig.add_subplot(122)

    ax1.plot(data_orbit.ra, data_orbit.dec, c='b')
    ax1.plot(data_orbit_rev.ra, data_orbit_rev.dec, c='r')
    ax1.scatter(ra, dec, c='k')

    ax1.set_xlim(0, 360)
    ax1.set_ylim(-90, 90)
    ax1.set_xlabel('l')
    ax1.set_ylabel('b')

    ax2.plot(data_orbit.ra, data_orbit.d, c='b')
    ax2.plot(data_orbit_rev.ra, data_orbit_rev.d, c='r')
    ax2.scatter(ra, d, c='k')

    ax2.set_xlim(0, 360)
    ax2.set_xlabel('l')
    ax2.set_ylabel('d (helio)')

    plt.show()

'''
================================================================================
 TESTING
================================================================================
'''

def test():
    test_o_l = 90
    test_o_b = 20
    test_o_d = 25
    test_o_vx = 150
    test_o_vy = -200
    test_o_vz = 100

    ts = np.linspace(0, 0.25, 1000)*u.Gyr
    sample = np.array([100, 250, 400, 500, 600, 750, 850])

    o = make_orbit([test_o_l, test_o_b, test_o_d, test_o_vx, test_o_vy, test_o_vz], pot=mwahpy_default_pot)

    l = np.take(o.ll(ts), sample)
    b = np.take(o.bb(ts), sample)
    d = np.take(o.dist(ts), sample)

    plt.scatter(l, b, s=5)
    plt.scatter(o.ll(ts), o.bb(ts), s=1)
    plt.show()

    plt.scatter(l, d, s=5)
    plt.scatter(o.ll(ts), o.dist(ts), s=1)
    plt.show()

    b_err = np.zeros(len(b)) + 0.05
    d_err = np.zeros(len(d)) + 1

    test_orbit_data = OrbitData(l, b, d, None, None, None, None, b_err, d_err, None, None, None, None)

    print('Goodness-of-Fit of actual values:')
    print(chi_squared([test_o_l, test_o_b, test_o_d, test_o_vx, test_o_vy, test_o_vz], data=test_orbit_data, pot=mwahpy_default_pot))

    params, x2 = fit_orbit(l, b, b_err, d, d_err, pot=mwahpy_default_pot)
    plot_orbit_gal(l, b, d, params, pot=mwahpy_default_pot)
    plot_orbit_icrs(l, b, d, params, pot=mwahpy_default_pot)

'''
================================================================================
 RUNTIME
================================================================================
'''

if __name__ == "__main__":
    test()



'''
The contents of this file are focused on testing the verious classes and routines
that are written out in the other files. All unit tests should be implemented
in this file so only one file needs to be run in order to complete all tests.
'''

import unittest
import numpy as np
import mwahpy.output_handler as oh
from mwahpy.timestep import Timestep
import mwahpy.coords as co

#===============================================================================
# UNIT TESTING
#===============================================================================
#Most of the tests in this file read in the test_bodies.out MW@h output file that
#is packaged in the tests folder along with this file. Without the test_bodies file,
#these tests will fail. If a different file is used, these tests will also fail.

#TODO: Complete tests for all functions in class

#NOTE: test_bodies.out does not have correct COM information, since it is a
#truncated version of a much larger MW@h output file.

#to run tests, run this file like you would run a regular python script

#to add more tests,
# 1) Find a test class that fits the description of the test you want to add,
#    then add that test to the class as a function beginning with "test", OR
# 2) Create a new class beginning with "test" and composed of just (unittest.TestCase)
#    and then add new functions that begin with "test" to that class.
#    This should be used if new tests do not fit the descriptions of other
#    test classes

#after changing any mwahpy file, this file should be run in order to make sure
#that functionality is still as is intended
#if tests don't pass, congratulations you broke something
#please fix it

prec = 8 #number of digits to round to when comparing floats
#WARNING: Tests may (but are not expected to) fail at high levels of prec

class TestTimestepClass(unittest.TestCase):

    def test_timestep_initialize(self):
        t = Timestep() #check default initialization
        t = oh.read_output('test_bodies.out') #check loading from a file

    def test_timestep_dict(self):
        t = oh.read_output('test_bodies.out')
        self.assertTrue(t.x[0] == t['x'][0])

        t.x[0] = 1
        self.assertTrue(t['x'][0] == 1)

    def test_timestep_iter(self):
        t = oh.read_output('test_bodies.out')

        for key, k in zip(t, t.index_list):
            self.assertTrue(t[key][0] == t[k][0])
            t[key][0] = 1

        self.assertTrue(t.x[0] == t['x'][0] == 1)

        for key in t:
            t[key] = np.append(t[key], 0)

        self.assertTrue(t.x[-1] == t['x'][-1] == 0)

    def test_copy(self):
        t = oh.read_output('test_bodies.out')

        t2 = t.copy()
        test = t.x[0]
        t.x[0] = 0

        self.assertTrue(t.x[0] != t2.x[0])
        self.assertTrue(t2.x[0] == test)

    def test_append_point(self):
        t = oh.read_output('test_bodies.out')

        t2 = t.copy()

        t2.append_point(t, 5)

        self.assertTrue(t.x[5] == t2.x[-1])
        self.assertTrue(len(t2) == len(t)+1)

    def test_split(self):
        t = oh.read_output('test_bodies.out')

        t1, t2 = t.split(5)
        self.assertTrue(t1.x[0] == t.x[0])
        self.assertTrue(t2.x[0] == t.x[5])

    def test_subset_rect(self):
        t = oh.read_output('test_bodies.out')
        tc = t.copy()

        tc.subset_rect(('x',), ((-1,1),))

        self.assertTrue(len(tc) == 7)

    def test_calcs(self): #makes sure that the values that
        #are not initially calculated on instantiation then run when the user
        #tries to get those values

        t = oh.read_output('test_bodies.out')

        self.assertTrue(len(t.pmra) == len(t))
        self.assertTrue(len(t.energy) == len(t))

class TestTimestepMethods(unittest.TestCase):

    def test_update(self):
        t = oh.read_output('test_bodies.out')
        t.update(force=True)

        #these values for the COM's are hard coded, unfortunately. But they are correct.
        self.assertTrue(len(t.center_of_mass) == 3)
        self.assertTrue(round(abs(t.center_of_mass[0] - 0.7835947518080879) + abs(t.center_of_mass[1] - 0.2947546230649471) + abs(t.center_of_mass[2] + 0.1318053758650839), prec) == 0)

        self.assertTrue(len(t.center_of_momentum) == 3)
        self.assertTrue(round(abs(t.center_of_momentum[0] - 6.001115213580641) + abs(t.center_of_momentum[1] + 65.29652414026405) + abs(t.center_of_momentum[2] + 26.462554427407223), prec) == 0)

        self.assertTrue(len(t.dist_from_com) == len(t))
        self.assertTrue(round(abs(t.dist_from_com[0] - ((t.center_of_mass[0] - t.x[0])**2 + (t.center_of_mass[1] - t.y[0])**2 + (t.center_of_mass[2] - t.z[0])**2)**0.5), prec) == 0)

class TestCoordsInverses(unittest.TestCase):

    #First order coordinate (position) transformations

    def test_cart_to_gal_rh(self):
        #right-handed
        x, y, z = 1., 2., 3.
        l, b, r = co.cart_to_gal(x, y, z)
        new_x, new_y, new_z = co.gal_to_cart(l, b, r)
        self.assertTrue((round(new_x, prec) == round(x, prec)) and (round(new_y, prec) == round(y, prec)) and (round(new_z, prec) == round(z, prec)))

    def test_cart_to_gal_lh(self):
        #left-handed
        x, y, z = 1., 2., 3.
        l, b, r = co.cart_to_gal(x, y, z, left_handed=True)
        new_x, new_y, new_z = co.gal_to_cart(l, b, r, left_handed=True)
        self.assertTrue((round(new_x, prec) == round(x, prec)) and (round(new_y, prec) == round(y, prec)) and (round(new_z, prec) == round(z, prec)))

    def test_cart_to_cyl(self):
        x, y, z = 1., 2., 3.
        R, z, phi = co.cart_to_cyl(x, y, z)
        new_x, new_y, new_z = co.cyl_to_cart(R, z, phi)
        self.assertTrue((round(new_x, prec) == round(x, prec)) and (round(new_y, prec) == round(y, prec)) and (round(new_z, prec) == round(z, prec)))

    def test_cart_to_sph(self):
        x, y, z = 1., 2., 3.
        phi, theta, r = co.cart_to_sph(x, y, z)
        new_x, new_y, new_z = co.sph_to_cart(phi, theta, r)
        self.assertTrue((round(new_x, prec) == round(x, prec)) and (round(new_y, prec) == round(y, prec)) and (round(new_z, prec) == round(z, prec)))

    #Second order coordinate (position) transformations

    def test_cyl_to_gal(self):
        R, z, phi = 10., 5., 120.
        l, b, r = co.cyl_to_gal(R, z, phi)
        new_R, new_z, new_phi = co.gal_to_cyl(l, b, r)
        self.assertTrue((round(new_R, prec) == round(R, prec)) and (round(new_z, prec) == round(z, prec)) and (round(new_phi, prec) == round(phi, prec)))

    #higher order coordinate transformations

    def test_sky_to_pole_null(self):
        tol = 1e-6 #tolerance for this test
        ra = np.linspace(0, 360, 10)
        dec = np.linspace(-90, 90, 10)
        ra_new, dec_new = co.sky_to_pole(ra, dec, (0, 90), (0, 0), wrap=True) #null transformation
        self.assertTrue((np.sum(np.abs(ra_new - ra)) <= tol) and (np.sum(np.abs(dec_new - dec)) <= tol))

    def test_sky_to_pole_inverse(self):
        tol = 1e-6 #tolerance for this test
        ra = np.linspace(10, 350, 10) #exact poles don't transform well
        dec = np.linspace(-80, 80, 10)
        L, B = co.sky_to_pole(ra, dec, (0, 0), (90, 0), wrap=True)
        ra_new, dec_new = co.sky_to_pole(L, B, (90, 0), (0, 90), wrap=True) #I'm like 75% sure this is the inverse transformation
        self.assertTrue((np.sum(np.abs(ra_new - ra)) <= tol) and (np.sum(np.abs(dec_new - dec)) <= tol))

#===============================================================================
# RUNTIME
#===============================================================================

if __name__ == '__main__':
    unittest.main()



'''
This file is a place to store flags that affect the operation of mwahpy
'''

#===============================================================================
# CLASSES
#===============================================================================

#allows for things like switching flags globally
class Flag():

    #should just be used within this file
    def __init__(self, default):
        self.val = default

    #allows you to use "if myflag:" to test for whether the flag is set
    def __bool__(self):
        return self.val

    #can externally switch flags on or off
    def switch(self, val):
        self.val = val

#===============================================================================
# FUNCTIONS
#===============================================================================

#checks if the user is running mwahpy in a jupyter notebook
#borrowed from https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook
#(thanks Gustavo Bezerra!)
def isnotebook():
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False      # Probably standard Python interpreter

#===============================================================================
# FLAGS
#===============================================================================

verbose = Flag(True) #If on, functions are a lot noisier
#will slow things down slightly, but it may be helpful for debugging

progress_bars = Flag(not(isnotebook())) #Display progress bars, on by default
#slows down a program somewhat, but is useful for making sure that
#a long program is running, as well as for debugging

#automatically turns progress bars off if the user is running a jupyter notebook,
#since the current implementation of progres bars clog up the client
#and lag out jupyter notebook

auto_update = Flag(True) #update all Timestep objects immediately when necessary
#this will keep things like center of mass, center of momentum, etc.
#accurate for the Timestep class, but will reduce performance somewhat.
#Should only be turned off if you REALLY know what you're doing, and you're sure
#that not updating Timestep objects won't impact what you're doing.

#Timestep classes can still be updated manually when necessary with <Timestep>.update()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mwahpy",
    version="1.4.4",
    author="Tom Donlon",
    author_email="donlot@rpi.edu",
    description="A python package for easily parsing and processing data from MilkyWay@home",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/thomasdonlon/mwahpy",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6.0',
    install_requires=[
        "numpy",
        "galpy",
        "matplotlib",
        "astropy",
        "scipy"
        ],
)



'''
The contents of this file are focused on the Timestep class, which is used for storage of
imported data from N-body output files.
'''

#TODO: UNIT TESTS
#TODO: Create sort function that sorts based on id (not really importamt)

#===============================================================================
# IMPORTS
#===============================================================================

#external imports
import numpy as np
from astropy.coordinates import SkyCoord
import astropy.units as u
import random
import galpy.potential

#mwahpy imports
from .mwahpy_glob import struct_to_sol, progress_bar
from .flags import auto_update, verbose
from .plot import scatter, trace_particle, hist, hist2d, fancy_plot
from .pot import mwahpy_default_pot, energy_offset
from .coords import get_rvpm

#===============================================================================
# CONSTANTS
#===============================================================================

basic_quantity_list = ['msol', 'l', 'b', 'ra', 'dec', 'phi', 'theta', 'dist', 'lx', 'ly', 'lz', 'lperp', 'ltot', 'r', 'R', 'vlos', 'vgsr', 'vrad', 'vrot', 'vR', 'dist_from_com']
rvpm_quantity_list = ['pmra', 'pmdec',  'pmtot', 'vtan']
energy_quantity_list = ['PE', 'KE', 'energy']

#===============================================================================
# TIMESTEP CLASS
#===============================================================================

#AttrDict is used as a helper class in Timestep to allow referencing attributes
#as dict keys and vice-versa.
#this is probably a bad way to implement this but it works, and it's better than
#making Timestep inherit from dict, which was the other solution I was able to strum up
class AttrDict(dict):
    def __init__(self):
        super(AttrDict, self).__init__()
        self.__dict__ = self

class Timestep():

    def __init__(self, typ=list(), id_val=list(), x=list(), y=list(), z=list(), vx=list(), vy=list(), vz=list(), mass=list(), center_of_mass=[0, 0, 0], center_of_momentum=[0, 0, 0], potential=None, time=None, nbody=None):
        #all this is typically specified by the read_output function in output_handler
        #read_output is the preferred way to input data to this data structure
        #but if you're really feeling adventurous you can always do it yourself

        #-----------------------------------------------------------------------
        # HOUSEKEEPING
        #-----------------------------------------------------------------------

        #this Timestep structure allows access to a dictionary like an attribute, i.e.
        #   d = AttrDict({'a':1, 'b':2})
        #   d['a'] == d.a //returns True
        #it seems like magic but it works by accessing the way python handles attributes
        #within each class. There are pros and cons, the most notable thing is
        #that it creates a memory leak in python <3.2.3
        ad = AttrDict() #this is a private helper class that allows for the
                        #desired indexing behavior
        self.__dict__ = ad.__dict__

        #-----------------------------------------------------------------------

        self.typ = np.array(typ).astype('int')   #0 if baryon, 1 if DM
        self.id = np.array(id_val).astype('int') #does not need to be unique ID
        self.x = np.array(x)       #Galactocentric Cartesian positions
        self.y = np.array(y)
        self.z = np.array(z)
        self.vx = np.array(vx)     #Galactocentric Cartesian velocities
        self.vy = np.array(vy)
        self.vz = np.array(vz)
        self.mass = np.array(mass) #structure masses

        #NOTE: Any time you update the position data, you have to update
        #   the center of mass (and same for velocity and COMomentum)
        #   This is done automatically if auto_update is on
        #   If you are manually screwing around with the provided values,
        #   you will need to update it manually
        self.center_of_mass = center_of_mass
        self.center_of_momentum = center_of_momentum

        #-----------------------------------------------------------------------
        # HOUSEKEEPING
        #-----------------------------------------------------------------------

        #if the time in Gyr is known for this simulation (i.e. is provided),
        #it is saved here.
        self.time = time

        #if the Timestep is part of a Nbody object, it saves that information here
        self.nbody = nbody

        #just in case the user wants to specify a different potential than the
        #mwahpy default, this is included
        self.potential = potential

        #this has to be manually updated any time a new iterable quantity is added
        #to the Timestep class. This allows us to control what values are iterated over.
        self.index_list = ['typ', 'id', 'x', 'y', 'z', 'vx', 'vy', 'vz', 'mass']
        self.index = 0

        #these should initially be set to false. If the user tries to get a value
        #that isn't yet calculated, then the getter calculates it. There's a
        #gigantic overhead on the rv/pm calculation and a pretty big overhead on
        #the energy calculation, so we avoid it until the user asks for those values.
        #The basic values are included here so that updating these values are easy
        self.have_basic = False
        self.have_rvpm = False
        self.have_energy = False

        #these flags should also initially be set to false. If the position or
        #velocity of one or more particles is changed at any point, the
        #corresponding flag should be swapped. This tells update() how much it
        #needs to update instead of running the whole routine every time, which
        #would be very time consuming.
        #these flags allow all calculated attributes to be updated whenever
        #changes are made, and at a reasonable speed.
        self.changed_pos = False
        self.changed_vel = False

    #---------------------------------------------------------------------------
    # METHODS
    #---------------------------------------------------------------------------

    #---------------------------------------------------------------------------
    # ITERATOR CONTROL
    #self.index_list allows us to ignore things that are not identically iterable in self.__dict__(),
    #such as the centers of mass and momentum

    def __iter__(self):
        self.index = 0
        return self

    def __next__(self):
        if self.index == len(self.index_list):
            raise StopIteration
        else:
            self.index += 1
            return self.index_list[self.index - 1]

    #---------------------------------------------------------------------------
    # These allow for the behavior we want from the class attributes, i.e. self.id == self['id']
    #__getattr__ and __getitem__ have qualifiers to make sure that loading rv/pm and energy
    #related values works as intended

    def __getattr__(self, i):
        if (not(self.have_basic) and i in basic_quantity_list):
            self.calc_basic()
        if (not(self.have_rvpm) and i in rvpm_quantity_list):
            self.calc_rvpm()
        if (not(self.have_energy) and i in energy_quantity_list):
            self.calc_energy()
        return self.__dict__[i]

    def __getitem__(self, i):
        if (not(self.have_basic) and i in basic_quantity_list):
            self.calc_basic()
        if (not(self.have_rvpm) and i in rvpm_quantity_list):
            self.calc_rvpm()
        if (not(self.have_energy) and i in energy_quantity_list):
            self.calc_energy(potential=self.potential)
        return self.__dict__[i]

    def __setitem__(self, i, val):
        self.__dict__[i] = val

        #set the flags that say we need to update the corresponding attributes
        #based on what was changed
        if i in ['x', 'y', 'z']:
            self.changed_pos = True
        elif i in ['vx', 'vy', 'vz']:
            self.changed_vel = True

    #---------------------------------------------------------------------------
    #a few routines that are important for the functionality of Timestep, but
    #don't really fit in elsewhere

    def __len__(self):
        return len(self.id)

    #replaces this Timestep with a blank Timestep
    def reset(self):
        self.typ = np.array([])
        self.id = np.array([])
        self.x = np.array([])
        self.y = np.array([])
        self.z = np.array([])
        self.vx = np.array([])
        self.vy = np.array([])
        self.vz = np.array([])
        self.mass = np.array([])

        self.center_of_mass = [0,0,0]
        self.center_of_momentum = [0,0,0]

        self.time = None
        self.nbody = None
        self.potential = None

        self.index_list = ['typ', 'id', 'x', 'y', 'z', 'vx', 'vy', 'vz', 'mass']
        self.index = 0

        self.have_basic = False
        self.have_rvpm = False
        self.have_energy = False

        self.changed_pos = False
        self.changed_vel = False

    def update(self, force=False):
        #only update the necessary attributes based on what changed
        #if force==True, then force updating everything. Should be only rarely needed
        if verbose:
            print('Updating data for new frame...')

        if force:
            self.changed_pos = True
            self.changed_vel = True

        if self.changed_pos:
            self.center_of_mass = [np.sum(self.x*self.mass/sum(self.mass)), np.sum(self.y*self.mass/sum(self.mass)), np.sum(self.z*self.mass/sum(self.mass))]
        if self.changed_vel:
            self.center_of_momentum = [np.sum(self.vx*self.mass/sum(self.mass)), np.sum(self.vy*self.mass/sum(self.mass)), np.sum(self.vz*self.mass/sum(self.mass))]

        if self.changed_pos or self.changed_vel:
            if self.have_basic:
                self.calc_basic()
            if self.have_rvpm:
                self.calc_rvpm()
            if self.have_energy:
                self.calc_energy(potential=self.potential)

        #the positions and velocities have not changed since updating
        #(since we just updated)
        changed_pos = False
        changed_vel = False

        if verbose:
            print('Done updating')

    #creates a deep copy of the Timestep object
    #this can't be done by iterating over the object, since comass etc. have to be copied as well
    def copy(self):
        out = Timestep()
        if self.have_basic: #these set the proper flags and index_list, etc. but don't cost computation time
            out.calc_basic()
        if self.have_rvpm:
            out.calc_rvpm()
        if self.have_energy:
            out.calc_energy()
        for key in self.__dict__.keys():
            if type(out[str(key)]) == type(np.array([])) or type(out[str(key)]) == type([]):
                out[str(key)] = self[str(key)].copy()
            else:
                out[str(key)] = self[str(key)]

        return out

    #---------------------------------------------------------------------------

    #the following methods are used to calculate values only if the user asks for them
    #this dramatically increases the speed of the package if the user is only using
    #a few values that need to be calculated

    def calc_basic(self):
        #none of these calculations take very long, so it's fair to group them up
        #this is only taken out of the original initialization because then you can update them
        #if you end up changing the provided values

        #this could be split up even more if any of the component value calculations
        #ends up taking a significant amount of time, but I doubt that will ever be the case

        if verbose:
            print('Calculating basic values...')

        self.msol = self.mass * struct_to_sol

        #position information
        self.r = (self.x**2 + self.y**2 + self.z**2)**0.5
        self.dist = ((self.x + 8)**2 + self.y**2 + self.z**2)**0.5
        self.R = (self.x**2 + self.y**2)**0.5

        #galactic coordinate information
        self.l = np.arctan2(self.y, self.x + 8)*180/np.pi
        self.b = np.arcsin(self.z/self.dist)*180/np.pi

        #ICRS information
        c = SkyCoord(l=self.l*u.degree, b=self.b*u.degree, frame='galactic')
        c_trans = c.transform_to('icrs')
        self.ra = c_trans.ra.degree
        self.dec = c_trans.dec.degree

        #galactocentric spherical coordinates
        #l = b = theta = phi = 0, theta=pi at NGP, -pi at SGP, phi increases in the same direction as l
        self.phi = np.arctan2(self.y, self.x)*180/np.pi
        self.theta = np.arcsin(self.z/self.r)*180/np.pi

        #angular momentum information
        self.lx = self.y * self.vz - self.z * self.vy
        self.ly = self.z * self.vx - self.x * self.vz
        self.lz = self.x * self.vy - self.y * self.vx
        self.lperp = (self.lx**2 + self.ly**2)**0.5
        self.ltot = (self.lx**2 + self.ly**2 + self.lz**2)**0.5

        #velocity information
        self.vgsr = ((self.x+8)*self.vx + self.y*self.vy + self.x*self.vz)/self.dist
        self.vlos = self.vgsr - 10.1*np.cos(self.b*np.pi/180)*np.cos(self.l*np.pi/180) - 224*np.cos(self.b*np.pi/180)*np.sin(self.l*np.pi/180) - 6.7*np.sin(self.b*np.pi/180)
        self.vrad = (self.x*self.vx + self.y*self.vy + self.z*self.vz)/self.r
        self.vrot = self.lz/(self.x**2 + self.y**2)**0.5 #should have a factor of cos(theta) I think? Need to define what I mean exactly by this quantity
                                                         #should also make a polar velocity if I'm going to do that
        self.vR = (self.x*self.vx + self.y*self.vy)/self.R

        #relative information
        self.dist_from_com = ((self.x - self.center_of_mass[0])**2 + (self.y - self.center_of_mass[1])**2 + (self.z - self.center_of_mass[2])**2)**0.5

        if not(self.have_basic): #only run if first time running method, not if updating
            self.index_list = self.index_list + basic_quantity_list
        self.have_basic = True #make sure the getter doesn't try to run this again

    def calc_rvpm(self):
        #the biggest source of overhead in this class is get_rvpm
        #so, don't run it unless you have to.

        if verbose:
            print('Calculating proper motion values...')

        self.pmra, self.pmdec = get_rvpm(self.ra, self.dec, self.dist, self.vx, self.vy, self.vz)[1:]
        #already get rv from vlos, don't need to save it as something else
        self.pmtot = (self.pmra**2 + self.pmdec**2)**0.5
        #4.848e-6 is arcsec->rad, 3.086e16 is kpc->km, and 3.156e7 is sidereal yr -> seconds
        self.vtan = 4.74*self.dist*self.pmtot #eq. to self.r*np.tan(self.pmtot*4.848e-6) * 3.086e16 / 3.156e7

        if not(self.have_rvpm): #only run if first time running method, not if updating
            self.index_list = self.index_list + rvpm_quantity_list

        self.have_rvpm = True #make sure the getter doesn't try to run this again

    def calc_energy(self, potential=None):
        #calculating the energy of every particle can generate some overhead,
        #so I've quarantined it with a flag.

        if verbose:
            print('Calculating energy values...')

        if potential == None:
            potential = mwahpy_default_pot

        #in a logarithmic halo, the magnitude of the potential doesn't impact the result,
        #just the difference in potentials. So, you can specify a potential offset
        #to keep bound objects' total energy negative.
        PE = galpy.potential.evaluatePotentials(potential, self.R * u.kpc, self.z*u.kpc, ro=8., vo=220.) + energy_offset
        KE = 0.5*(self.vx**2 + self.vy**2 + self.vz**2)

        #set attributes
        self.PE = PE
        self.KE = KE
        self.energy = PE + KE

        #allow iteration over these attributes
        if not(self.have_energy):  #only run if first time running method, not if updating
            self.index_list = self.index_list + energy_quantity_list

        self.have_energy = True #make sure the getter doesn't try to run this again

    #---------------------------------------------------------------------------

    #cuts the first n entries from every attribute in the Timestep structure
    def cut_first_n(self, n):
        for key in self:
            self[key] = self[key][n:]
        if auto_update:
            self.update()

    #cuts the last n entries from every attribute in the Timestep structure
    def cut_last_n(self, n):
        l = len(self) #length changes during this, so have to save it
        for key in self:
            self[key] = self[key][:l-n]
        if auto_update:
            self.update()

    #cut the data to only include the values at the given indices
    def take(self, indices):
        #indices: the indices you want to take, must be array-like
        for key in self:
            self[key] = np.take(self[key], indices)
        if auto_update:
            self.update()

    #cut the data to only include particles with id in ids
    def take_by_id(self, ids):
        #ids: the ids of the particles you want to take, must be array-like
        id_list = self.id.tolist()
        indices_list = []
        for i in ids: #there must be a quicker way to do this, but I can't think of one right now
            indices_list.append(id_list.index(i))
        for key in self:
            self[key] = np.take(self[key], indices_list)
        if auto_update:
            self.update()

    #resets the IDs of the Timestep instance
    def reset_ids(self):
        self.id = np.arange(len(self))

    #splits the Timestep into two new Timestep structures,
    #the first has the data points up to entry n and
    #the second has the data points after entry n
    def split(self, n):
        Timestep1 = self.copy()
        Timestep2 = self.copy()

        Timestep1.cut_last_n(len(self) - n)
        Timestep2.cut_first_n(n)

        if auto_update:
            Timestep1.update()
            Timestep2.update()

        return Timestep1, Timestep2

    #splits the Timestep into a list of new Timestep structures,
    #where the Timestep is split every time ID wraps back to zero
    #TODO: the dwarf id start at 1, not 0
    def split_at_id_wrap(self):

        outlist = []
        indices = np.where(self.id==0)[0] #1D list of arrays

        Timestep2 = self.copy()
        i = 1
        while i < len(indices): #pseudo-recursive
            Timestep1, Timestep2 = Timestep2.split(indices[i] - indices[i-1])
            outlist.append(Timestep1)
            i += 1
        outlist.append(Timestep2)

        if auto_update:
            for t in outlist:
                t.update()

        return outlist

    #append a Timestep object onto this one
    def append_timestep(self, t):
        #t: the Timestep to append to this object
        for key, dkey in zip(self, t):
            self[key] = np.append(self[key], t[dkey])

        if auto_update:
            self.update()

    #append the nth item of another Timestep object onto this one
    #WARNING: Extremely time intensive if used repeatedly, np.append is not a speedy function
    #   you should always try to append a timestep instead if possible
    def append_point(self, t, n=0, id=None):
        #t: the Timestep object with the item being appended
        #n (optional): the index of the item in t to be appended
        #id (optional): if not None, uses finds the first item with matching id and appends that
        if id:
            n = np.where(t['id'] == id)[0]
        for key, tkey in zip(self, t):
            self[key] = np.append(self[key], t[tkey][n])

        if auto_update:
            self.update()

    #centers the timestep so that the COMs are now both zero vectors
    #this can be useful if running live simulations, which can pick up nonzero
    #   overall velocities when they relax from unstable virial equilibrium ICs
    #Conserves the physics of the system under a change of reference frame
    def recenter(self):
        if verbose:
            print('Recentering the data on the centers of mass and momentum...', end='')

        #center the positions of the particles
        self.x = self.x - self.center_of_mass[0]
        self.y = self.y - self.center_of_mass[1]
        self.z = self.z - self.center_of_mass[2]

        #center the velocities of the particles
        self.vx = self.vx - self.center_of_momentum[0]
        self.vy = self.vy - self.center_of_momentum[1]
        self.vz = self.vz - self.center_of_momentum[2]

        if verbose:
            print('done')

        if auto_update:
            self.update()

    #centers the timestep so that the COMs of all individual components are now zero vectors
    #this can be useful if running live simulations, which can pick up nonzero
    #   overall velocities when they relax from unstable virial equilibrium ICs
    #WARNING: Does not conserve the physics of the system. This is more for testing
    #   or debugging purposes, or if you really know what you're doing.
    def recenter_each_component(self):
        #center the positions of the particles
        data_list = self.split_at_id_wrap()

        for i in data_list:
            i.recenter()

        self.reset()
        for i in data_list:
            self.append_timestep(i)

        if auto_update:
            self.update()

    #cut the Timestep so only baryons (typ==0) remain
    def only_baryons(self):
        #find indices of particles with typ==0
        ind = []
        for i, t in zip(range(len(self.typ)), self.typ): #ugly but fast I think
            if t==0:
                ind.append(i)

        #cut the data
        self.take(ind)

    #cut the Timestep so only dark matter particles (typ==1) remain
    def only_dark_matter(self):
        #find indices of particles with typ==1
        ind = []
        for i, t in zip(range(len(self.typ)), self.typ): #ugly but fast I think
            if t==0:
                ind.append(i)

        #cut the data
        self.take(ind)

    #---------------------------------------------------------------------------
    # OUTPUT
    #---------------------------------------------------------------------------

    def print_particle(self, n, dec=8):
        #n (int): the location of the particle that you want the information for
        #TODO: allow for using the id to find the particle
        print('Printing data for Particle '+str(n)+':')

        outstr = '('
        for key in self:
            outstr = outstr + key + ':' + str(round(self[key][n],dec)) + ', '
        outstr = outstr + ')'
        print(outstr)

    #---------------------------------------------------------------------------
    # PLOTTING
    #---------------------------------------------------------------------------
    #NOTE: the actual plotting routines should be implemented in plot.py
    #this is just to allow you to access the routines as timestep.<method>()

    def scatter(self, x, y, **kwargs):
        scatter(self, x, y, **kwargs)

    def trace_particle(self, id, x, y, **kwargs):
        trace_particle(self, id, x, y, **kwargs)

    def hist(self, x, **kwargs):
        hist(self, x, **kwargs)

    def hist2d(self, x, y, **kwargs):
        hist2d(self, x, y, **kwargs)

    #uses kwargs for plt.plot()
    def fancy_plot(self, x, y, **kwargs):
        fancy_plot(self, x, y, **kwargs)

    #---------------------------------------------------------------------------
    # MISCELLANEOUS
    #---------------------------------------------------------------------------

    #make an n-dimensional rectangular cut on the data
    #TODO: make an inverted method, i.e. cut out things within the bounds
    def subset_rect(self, axes, bounds):
        #axes ([str, ...]): the parameters/values that you are cutting on. Input as a list of strings
        #   The strings must be in self.index_list
        #bounds ([(float, float), ...]): The boundary conditions that you are cutting on
        #   given as a list of tuples, each with two floats.
        #   If you wish to only declare a minimum or maximum, then leave the other value as None
        #
        #Both axes and bounds can be given as a tuple instead of a list
        #
        #This is substantially faster than the previous subset function, which performed like 30
        #np.appends for every index that fit the criteria in each axis.

        if type(axes) not in [type([]), type((1,1))]: #make the input compatible if it is only 1 axis
            raise Exception('Axes must be a list or a tuple, but was of type ' + str(type(axes)))

        if len(axes) != len(bounds): #need the same number of axes and bounds
            raise Exception('Number of axes is ' + str(len(axes)) + ', but number of bounds is ' + str(len(bounds)))

        #make the first cut, gives us some place to start on for the indices
        if bounds[0][0] > bounds[0][1]: #make sure the bounds are input correctly
            raise Exception('First value in bound was larger than second value')

        indices = np.intersect1d(np.where(self[axes[0]] > bounds[0][0]), np.where(self[axes[0]] < bounds[0][1]))

        for a, b in zip(axes[1:], bounds[1:]): #already performed first cut
            if b[0] > b[1]: #make sure the bounds are input correctly
                raise Exception('First value in bound was larger than second value')

            #slowly wittle down the number of indices that fit the criteria
            indices = np.intersect1d(indices, np.intersect1d(np.where(self[a] > b[0]), np.where(self[a] < b[1])))

        #cut the sample down to the indices that work
        self.take(indices)
        self.update()

    #TODO: make n-dimensional like subset_rect
    #make a circular cut of the data in 2 dimensions
    def subset_circ(self, axes, rads, centers):
        #axs ([str]): the axes to cut on
        #rads ([float]): The radiii of the circular cut (along each axis)
        #center ([float], len=#axes): The center around which to make the circular cut
        #
        #This is substantially faster than the previous subset function, which performed like 30
        #np.appends for every index that fit the criteria in each axis.

        #get the indices that lie within the cut
        dist = 0
        for a, r, c in zip(axes, rads, centers):
            dist += (self[a] - c)**2/r**2

        indices = np.where(dist < 1)[0]

        #cut the sample down to the indices that work
        self.take(indices)
        self.update()

    #cut the Timestep instance to n random stars from the Timestep instance
    #uses the reservoir algorithm for single-pass random sampling
    def rand_sample(self, n):
        #n: the number of stars to sample

        reservoir = np.arange(0,n)
        for i in np.arange(n,len(self)-1):
            r = random.randint(0,len(self)-1)
            if r < n:
                reservoir[r] = i

        self.take(reservoir.astype('int'))
        self.update()

    #cut the Timestep instance to every nth star from the Timestep instance
    def subsample(self, n, offset=0):
        #n: take the star every n rows (n should be integer >= 1)

        i = offset
        indices = []
        while i < len(self):
            indices.append(i)
            i += n

        self.take(indices)
        self.update()

    #rotate all particles ccw around the Z axis by theta radians
    def rot_around_z_axis(self, theta):
        #theta: rotation angle in radians

        newx = self.x * np.cos(theta) - self.y * np.sin(theta)
        newy = self.x * np.sin(theta) + self.y * np.cos(theta)
        newvx = self.vx * np.cos(theta) - self.vy * np.sin(theta)
        newvy = self.vx * np.sin(theta) + self.vy * np.cos(theta)

        self.x = newx
        self.y = newy
        self.vx = newvx
        self.vy = newvy

        self.update()

#===============================================================================
# FUNCTIONS INVOLVING TIMESTEP CLASSES
#===============================================================================

#Timestep -> np array of floats
#computes the energy of each particle based only on the self-gravity
#of the particles and their velocities w.r.t. the COM's of the particles.
#slow because I am just implementing straight N^2 gravity calcs, but
#not sure if there's a better way to do this
#NOTE: This is energy per unit mass, not energy
#WARNING: Only returns initial energy of the particles in a dwarf when run
#         on a Timestep where the entire dwarf is still bound (e.g. the
#         beginning of a simulation), otherwise output is ~meaningless
def get_self_energies(t):
    #t: the Timestep

    gc = 4.30091e-3 * struct_to_sol / 1000 #Newton's gravitational constant in units of kpc/struct mass * (km/s)^2

    t = t.copy() #don't hurt the poor innocent Timestep
    t.recenter() #recenters both positions and velocities

    pot_energies = []

    #compute the self gravitational potential energy of each particle
    i = 0 #tracking
    for x1, y1, z1 in zip(t.x, t.y, t.z):
        progress_bar(i, len(t.x))
        i += 1

        r = ((x1 - t.x)**2 + (y1 - t.y)**2 + (z1 - t.z)**2)**0.5
        pe = np.sum(-1 * gc * t.mass[r != 0] / r[r != 0]) #potential energy is negative, remove where r == 0
        pot_energies.append(pe)

    pot_energies = np.array(pot_energies) #do it this way because np.append is horrendously slow

    #compute the kinetic energy of each particle (within reference frame of
    #the COM's of the particles)
    kin_energies = 0.5*(t.vx**2 + t.vy**2 + t.vz**2)

    energies = pot_energies + kin_energies

    return energies

#calculates where the progenitor of a stream is
#More specifically, it computes where the highest concentration of baryons is
#in Galactic Cartesian coordinates
#written by Eric Mendelsohn, 2021
#adapted for mwahpy by Tom Donlon, 2021

#TODO: Can have trouble finding progenitors around poles. If progenitor is
#   above 60 deg or so, could apply arbitrary spherical rotation to bring it back
#   into a spot that the algorithm handles better
def find_progenitor(t, ran=100.):
    #t (Timestep): the Timestep that you are running the algorithm on
    #ran (float, optional): the maximum distance range to search, in kpc

    bins = 1000

    #don't wreck the data that's passed in
    t = t.copy()

    #cut the Timestep so only the baryons are considered
    t.only_baryons()

    #cut the data to the prescribed distance ranges
    t.subset_rect(['r'], [(0.0001, ran)])

    #histogram the data in spherical galactic coords
    h = np.histogram2d(t.theta, t.phi, bins=bins, range=((-90, 90), (-180, 180)))[0]
    phi_vals = np.linspace(-180, 180, bins)
    theta_vals = np.linspace(-90, 90, bins)

    #Find (theta,phi) with the maximum star density
    theta_max_bin = 0
    phi_max_bin = 0
    for i in range(len(h)):
        for j in range(len(h[i])):
            if (h[theta_max_bin][phi_max_bin]/np.cos(theta_vals[theta_max_bin]*np.pi/180) < h[i][j]/np.cos(theta_vals[i]*np.pi/180)):
                theta_max_bin = i
                phi_max_bin = j

    #Pull out stars in max bin
    theta_binsize = 180/bins
    phi_binsize = 360/bins
    accepted_stars = t.copy()
    accepted_stars.subset_rect(['theta', 'phi'], [(theta_vals[theta_max_bin]-theta_binsize/2, theta_vals[theta_max_bin]+theta_binsize/2), \
                                                             (phi_vals[phi_max_bin]-phi_binsize/2, phi_vals[phi_max_bin]+phi_binsize/2)])

    #Find r of progenitor while correcting for sigma clipping
    for n in range(50):
        #Find average and standard deviation of r
        star_num = len(accepted_stars)
        if (star_num < 2):
            print('COULD NOT RESOLVE PROGENITOR')
            return [0.0,0.0,0.0]
        r_total = np.sum(accepted_stars.r)
        r2_total = np.sum((accepted_stars.r)**2)
        r_mean = r_total/star_num
        r2_mean = r2_total/star_num
        r_sig = ((r2_mean - r_mean**2.0)*star_num/(star_num - 1.0))**0.5

        #Perform 2.5-sigma clipping
        new_accepted_stars = []
        for i in range(len(accepted_stars)):
            if(abs(accepted_stars.r[i] - r_mean)/abs(r_sig) < 2.5):
                new_accepted_stars.append(i)
        accepted_stars.take(new_accepted_stars)

    #Report coordinates in XYZ
    x_val = r_mean*np.cos(theta_vals[theta_max_bin]*np.pi/180)*np.cos(phi_vals[phi_max_bin]*np.pi/180)
    y_val = r_mean*np.cos(theta_vals[theta_max_bin]*np.pi/180)*np.sin(phi_vals[phi_max_bin]*np.pi/180)
    z_val = r_mean*np.sin(theta_vals[theta_max_bin]*np.pi/180)
    pos = [x_val,y_val,z_val]

    return pos



'''
This file is a place to store constants and other information
that are widely used throughout mwahpy.
'''

#TODO: Get more accurate solar velocity and position of Sun from GC
#TODO: Make more tests and put them all in a single test.py file
#TODO: Ensure all imports are actually used (and necessary) for each file

#===============================================================================
# IMPORTS
#===============================================================================

import astropy.units as u

#===============================================================================
# CLASSES
#===============================================================================

#currently just used for the position of the Sun, but could be more useful in other places
class PhaseSpace():

    def __init__(self, ps):
        self.x = ps[0]
        self.y = ps[1]
        self.z = ps[2]
        self.vx = ps[3]
        self.vy = ps[4]
        self.vz = ps[5]

    #this way you can change the phase space used for the Sun within the entire package
    def update(self, ps):
        self.x = ps[0]
        self.y = ps[1]
        self.z = ps[2]
        self.vx = ps[3]
        self.vy = ps[4]
        self.vz = ps[5]

#===============================================================================
# CONSTANTS
#===============================================================================

G = 6.674e-11*u.m**3/(u.kg*u.s**2)

struct_to_sol = 222288.47 #this many solar masses make up one structural nass unit (the output of mwah)

kms_to_kpcgyr = 1.023 #1 km/s is 1.023 kpc/Gyrb

kpcgyr_to_kms = 0.978 #1 kpc/Gyr is 0.978 km/s

#define transformations from heliocentric -> galactocentric frame
#by defining heliocentric frame here

#right-handed coordinate frame(!)
#Hogg et al. (2005)
solar_ps = PhaseSpace([-8., 0., 0., 10.1, 224.0, 6.7])

#===============================================================================
# FUNCTIONS
#===============================================================================

#prints out a progress bar in the terminal
#"borrowed" from https://stackoverflow.com/questions/6169217/replace-console-output-in-python
def progress_bar(value, endvalue, bar_length=20):

    percent = float(value) / endvalue
    arrow = '-' * int(round(percent * bar_length)-1) + '>'
    spaces = ' ' * (bar_length - len(arrow))

    print('\r[{0}] {1}%'.format(arrow + spaces, int(round(percent * 100))), end='')

#get length of a file (in lines), given the filename
def file_len(f):
    with open(f) as f:
        for i, l in enumerate(f):
            pass
    return i + 1
